"""Base module for interacting with LLM APIs."""
import re
import os
import json
import time
import datetime
from typing import Any, AsyncGenerator, List, Dict, Optional, Union, Generator, Tuple
from attr import dataclass

import openai
from pydantic import BaseModel
from dotenv import load_dotenv
from litellm import completion, acompletion
from litellm import ModelResponse

from promptmodel.utils.types import LLMResponse, LLMStreamResponse
from promptmodel.utils import logger
from promptmodel.utils.enums import ParsingType, ParsingPattern, get_pattern_by_type
from promptmodel.utils.output_utils import convert_str_to_type, update_dict
from promptmodel.utils.prompt_util import (
    num_tokens_for_messages,
    num_tokens_from_function_call_output,
    num_tokens_from_functions_input,
)

load_dotenv()


class Role:
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    FUNCTION = "function"


class OpenAIMessage(BaseModel):
    role: Optional[str] = None
    content: Optional[str] = ""
    function_call: Optional[Dict[str, Any]] = None
    name: Optional[str] = None


DEFAULT_MODEL = "gpt-3.5-turbo"


@dataclass
class ParseResult:
    parsed_outputs: Dict[str, Any]
    error: bool
    error_log: Optional[str]


class LLM:
    def __init__(self, rate_limit_manager=None):
        self._rate_limit_manager = rate_limit_manager

    @classmethod
    def __parse_output_pattern__(
        cls, raw_output: str, parsing_type: Optional[ParsingType] = None
    ) -> ParseResult:
        if parsing_type is None:
            return None
        parsing_pattern = get_pattern_by_type(parsing_type)
        whole_pattern = parsing_pattern["whole"]
        parsed_results = re.findall(whole_pattern, raw_output, flags=re.DOTALL)
        parsed_outputs = {}
        error = False
        error_log: str = None

        try:
            for parsed_result in parsed_results:
                key = parsed_result[0]
                type_str = parsed_result[1]
                value = convert_str_to_type(parsed_result[2], type_str)
                parsed_outputs[key] = value
        except Exception as error:
            error = True
            error_log = str(error)

        return ParseResult(
            parsed_outputs=parsed_outputs,
            error=error,
            error_log=error_log,
        )

    def __validate_openai_messages(
        self, messages: List[Dict[str, str]]
    ) -> List[OpenAIMessage]:
        """Validate and convert list of dictionaries to list of OpenAIMessage."""
        res = []
        for message in messages:
            res.append(OpenAIMessage(**message))
        return res

    def run(
        self,
        messages: List[Dict[str, str]],
        functions: List[Any] = [],
        model: Optional[str] = DEFAULT_MODEL,
        *args,
        **kwargs,
    ):
        """Return the response from openai chat completion."""
        response = None
        try:
            response = completion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                functions=functions,
            )

            content = (
                response.choices[0]["message"]["content"]
                if "content" in response.choices[0]["message"]
                else None
            )
            call_func = (
                response.choices[0]["message"]["function_call"]
                if "function_call" in response.choices[0]["message"]
                else None
            )
            return LLMResponse(
                api_response=response, raw_output=content, function_call=call_func
            )
        except Exception as e:
            if response:
                return LLMResponse(api_response=response, error=True, error_log=str(e))
            else:
                return LLMResponse(api_response=None, error=True, error_log=str(e))

    async def arun(
        self,
        messages: List[Dict[str, str]],
        functions: List[Any] = [],
        model: Optional[str] = DEFAULT_MODEL,
        *args,
        **kwargs,
    ):
        """Return the response from openai chat completion."""
        response = None
        try:
            response = await acompletion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                functions=functions,
            )
            content = (
                response.choices[0]["message"]["content"]
                if "content" in response.choices[0]["message"]
                else None
            )
            call_func = (
                response.choices[0]["message"]["function_call"]
                if "function_call" in response.choices[0]["message"]
                else None
            )
            return LLMResponse(
                api_response=response, raw_output=content, function_call=call_func
            )
        except Exception as e:
            if response:
                return LLMResponse(api_response=response, error=True, error_log=str(e))
            else:
                return LLMResponse(api_response=None, error=True, error_log=str(e))

    def stream(
        self,
        messages: List[Dict[str, str]],  # input
        functions: List[Any] = [],
        model: Optional[str] = DEFAULT_MODEL,
        *args,
        **kwargs,
    ):
        """Stream openai chat completion."""
        response = None
        try:
            # load_prompt()
            start_time = datetime.datetime.now()
            response = completion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                stream=True,
                functions=functions,
            )
            
            for result in self.__llm_stream_response_generator__(
                    messages, response, start_time, functions
                ):
                    yield result
        except Exception as e:
            return LLMStreamResponse(error=True, error_log=str(e))

    async def astream(
        self,
        messages: List[Dict[str, str]],
        functions: List[Any] = [],
        model: Optional[str] = DEFAULT_MODEL,
        *args,
        **kwargs,
    ) -> Generator[LLMStreamResponse, None, None]:
        """Parse & stream output from openai chat completion."""
        response = None
        try:
            start_time = datetime.datetime.now()
            response = await acompletion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                stream=True,
                functions=functions,
            )

            async for result in self.__llm_stream_response_agenerator__(
                    messages, response, start_time, functions
                ):
                    yield result
        except Exception as e:
            yield LLMStreamResponse(error=True, error_log=str(e))

    def run_and_parse(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: List[Any] = [],
        output_keys: Optional[List[str]] = None,
        model: Optional[str] = DEFAULT_MODEL,
    ) -> Dict[str, str]:
        """Parse and return output from openai chat completion."""
        response = None
        try:
            parse_success = True
            response = completion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                functions=functions,
            )
            raw_output = response.choices[0]["message"]["content"]

            parse_result: Optional[ParseResult] = self.__parse_output_pattern__(
                raw_output, parsing_type
            )
            error_log = parse_result.error_log

            call_func = (
                response.choices[0]["message"]["function_call"]
                if "function_call" in response.choices[0]["message"]
                else None
            )

            if (
                output_keys is not None
                and set(parse_result.parsed_outputs.keys()) != set(output_keys)
                or parse_result.error
            ):
                parse_success = False
                error_log = "Output keys do not match with parsed output keys"

            return LLMResponse(
                api_response=response,
                raw_output=raw_output,
                parsed_outputs=parse_result.parsed_outputs if parse_result else None,
                function_call=call_func,
                error=not parse_success,
                error_log=error_log,
            )
        except Exception as e:
            if response:
                return LLMResponse(api_response=response, error=True, error_log=str(e))
            else:
                return LLMResponse(api_response=None, error=True, error_log=str(e))

    async def arun_and_parse(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: List[Any] = [],
        output_keys: Optional[List[str]] = None,
        model: Optional[str] = DEFAULT_MODEL,
    ) -> Dict[str, str]:
        """Generate openai chat completion asynchronously, and parse the output.
        Example prompt is as follows:
        -----
        Given a topic, you are required to generate a story.
        You must follow the provided output format.

        Topic:
        {topic}

        Output format:
        [Story]
        ...
        [/Story]

        Now generate the output:
        """
        response = None
        parsed_success = True
        try:
            response = await acompletion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                functions=functions,
            )
            raw_output = response.choices[0]["message"]["content"]
            parse_result: ParseResult = self.__parse_output_pattern__(
                raw_output, parsing_type
            )
            error_log = parse_result.error_log

            call_func = (
                response.choices[0]["message"]["function_call"]
                if "function_call" in response.choices[0]["message"]
                else None
            )

            if (
                output_keys is not None
                and set(parse_result.parsed_outputs.keys()) != set(output_keys)
                or parse_result.error
            ):
                parse_success = False
                error_log = "Output keys do not match with parsed output keys"

            if (
                output_keys is not None
                and set(parse_result.parsed_outputs.keys()) != set(output_keys)
            ) and parse_success:
                parsed_success = False
                error_log = "Output keys do not match with parsed output keys"

            return LLMResponse(
                api_response=response,
                parsed_outputs=parse_result.parsed_outputs,
                function_call=call_func,
                error=not parsed_success,
                error_log=error_log,
            )
        except Exception as e:
            if response:
                return LLMResponse(api_response=response, error=True, error_log=str(e))
            else:
                return LLMResponse(api_response=None, error=True, error_log=str(e))

    def stream_and_parse(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: List[Any] = [],
        output_keys: Optional[List[str]] = None,
        model: Optional[str] = DEFAULT_MODEL,
        **kwargs,
    ) -> Generator[LLMStreamResponse, None, None]:
        """Parse & stream output from openai chat completion."""
        response = None
        try:
            if parsing_type == ParsingType.COLON.value:
                # cannot stream colon type
                yield LLMStreamResponse(
                    error=True, error_log="Cannot stream colon type"
                )
                return
            start_time = datetime.datetime.now()
            response = completion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                stream=True,
                functions=functions,
            )

            parsed_outputs = {}
            error_occurs = False
            error_log = ""

            if parsing_type is None:
                for result in self.__llm_stream_response_generator__(
                    messages, response, start_time, functions
                ):
                    yield result

                    if result.error and not error_occurs:
                        error_occurs = True
                        error_log = result.error_log

            elif parsing_type == ParsingType.DOUBLE_SQUARE_BRACKET.value:
                for result in self.__double_type_sp_generator__(
                    messages, response, parsing_type, start_time, functions
                ):
                    yield result
                    if result.parsed_outputs:
                        parsed_outputs = update_dict(
                            parsed_outputs, result.parsed_outputs
                        )
                    if result.error and not error_occurs:
                        error_occurs = True
                        error_log = result.error_log
            else:
                for result in self.__single_type_sp_generator__(
                    messages, response, parsing_type, start_time
                ):
                    yield result
                    if result.parsed_outputs:
                        parsed_outputs = update_dict(
                            parsed_outputs, result.parsed_outputs
                        )
                    if result.error and not error_occurs:
                        error_occurs = True
                        error_log = result.error_log

            if (
                output_keys is not None
                and set(parsed_outputs.keys()) != set(output_keys)
            ) and not error_occurs:
                error_occurs = True
                error_log = "Output keys do not match with parsed output keys"
                yield LLMStreamResponse(error=True, error_log=error_log)

        except Exception as e:
            return LLMStreamResponse(error=True, error_log=str(e))

    async def astream_and_parse(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: List[Any] = [],
        output_keys: Optional[List[str]] = None,
        model: Optional[str] = DEFAULT_MODEL,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        """Parse & stream output from openai chat completion."""
        response = None
        try:
            if parsing_type == ParsingType.COLON.value:
                # cannot stream colon type
                yield LLMStreamResponse(
                    error=True, error_log="Cannot stream colon type"
                )
                return
            start_time = datetime.datetime.now()
            response = await acompletion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                stream=True,
                functions=functions,
            )

            parsed_outputs = {}
            error_occurs = False
            if parsing_type is None:
                async for result in self.__llm_stream_response_agenerator__(
                    messages, response, start_time, functions
                ):
                    yield result

                    if result.error and not error_occurs:
                        error_occurs = True
                        error_log = result.error_log

            elif parsing_type == ParsingType.DOUBLE_SQUARE_BRACKET.value:
                async for result in self.__double_type_sp_agenerator__(
                    messages, response, parsing_type, start_time, functions
                ):
                    yield result
                    if result.parsed_outputs:
                        parsed_outputs = update_dict(
                            parsed_outputs, result.parsed_outputs
                        )
                    if result.error and not error_occurs:
                        error_occurs = True
            else:
                async for result in self.__single_type_sp_agenerator__(
                    messages, response, parsing_type, start_time, functions
                ):
                    yield result
                    if result.parsed_outputs:
                        parsed_outputs = update_dict(
                            parsed_outputs, result.parsed_outputs
                        )
                    if result.error and not error_occurs:
                        error_occurs = True

            if (
                output_keys is not None
                and set(parsed_outputs.keys()) != set(output_keys)
            ) and not error_occurs:
                error_occurs = True
                error_log = "Output keys do not match with parsed output keys"
                yield LLMStreamResponse(error=True, error_log=error_log)

        except Exception as e:
            yield LLMStreamResponse(error=True, error_log=str(e))

    def make_model_response(
        self,
        chunk: dict,
        response_ms,
        messages: List[Dict[str, str]],
        raw_output: str,
        function_list: List[Any] = [],
        function_call: Optional[dict] = None,
    ) -> ModelResponse:
        count_start_time = datetime.datetime.now()
        prompt_token: int = num_tokens_for_messages(
            messages=messages, model=chunk["model"]
        )
        completion_token: int = num_tokens_for_messages(
            model=chunk["model"],
            messages=[{"role": "assistant", "content": raw_output}],
        )

        if len(function_list) > 0:
            function_list_token = num_tokens_from_functions_input(
                functions=function_list, model=chunk["model"]
            )
            prompt_token += function_list_token

        if function_call:
            function_call_token = num_tokens_from_function_call_output(
                function_call_output=function_call, model=chunk["model"]
            )
            completion_token += function_call_token

        count_end_time = datetime.datetime.now()
        logger.debug(
            f"counting token time : {(count_end_time - count_start_time).total_seconds() * 1000} ms"
        )

        usage = {
            "prompt_tokens": prompt_token,
            "completion_tokens": completion_token,
            "total_tokens": prompt_token + completion_token,
        }
        res = ModelResponse(
            id=chunk["id"],
            created=chunk["created"],
            model=chunk["model"],
            usage=usage,
            response_ms=response_ms,
        )
        res["choices"][0]["finish_reason"] = chunk["choices"][0]["finish_reason"]
        res["choices"][0]["message"]["content"] = (
            raw_output if raw_output != "" else None
        )
        res["response_ms"] = response_ms
        if function_call:
            res.choices[0]["message"]["function_call"] = function_call
        return res

    def __double_type_sp_generator__(
        self,
        messages: List[Dict[str, str]],
        response: Generator,
        parsing_type: ParsingType,
        start_time: datetime.datetime,
        functions: List[Any] = [],
    ):
        try:
            parsing_pattern = get_pattern_by_type(parsing_type)
            start_tag = parsing_pattern["start"]
            start_fstring = parsing_pattern["start_fstring"]
            end_fstring = parsing_pattern["end_fstring"]
            start_token = parsing_pattern["start_token"]
            end_token = parsing_pattern["end_token"]

            buffer = ""
            raw_output = ""
            active_key = None
            stream_pause = False
            end_tag = None
            function_call = {"name": "", "arguments": ""}
            for chunk in response:
                if (
                    "content" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["content"] is not None
                ):
                    stream_value: str = chunk["choices"][0]["delta"]["content"]
                    raw_output += stream_value
                    yield LLMStreamResponse(raw_output=stream_value)
                    buffer += stream_value

                    while True:
                        if active_key is None:
                            keys = re.findall(start_tag, buffer, flags=re.DOTALL)
                            if len(keys) > 1:
                                yield LLMStreamResponse(
                                    error=True,
                                    error_log="Parsing error : Nested key detected",
                                )
                                break
                            if len(keys) == 0:
                                break  # no key
                            active_key, active_type = keys[0]
                            end_tag = end_fstring.format(key=active_key)
                            # delete start tag from buffer
                            start_pattern = start_fstring.format(
                                key=active_key, type=active_type
                            )
                            buffer = buffer.split(start_pattern)[-1]

                        else:
                            if (
                                stream_value.find(start_token) != -1
                            ):  # start token appers in chunk -> pause
                                stream_pause = True
                                break
                            elif stream_pause:
                                if (
                                    buffer.find(end_tag) != -1
                                ):  # if end tag appears in buffer
                                    yield LLMStreamResponse(
                                        parsed_outputs={
                                            active_key: buffer.split(end_tag)[0]
                                        }
                                    )
                                    buffer = buffer.split(end_tag)[-1]
                                    active_key = None
                                    stream_pause = False
                                elif (
                                    stream_value.find(end_token) != -1
                                ):  # if ("[blah]" != end_pattern) appeared in buffer
                                    if (
                                        buffer.find(end_token + end_token) != -1
                                    ):  # if ]] in buffer -> error
                                        yield LLMStreamResponse(
                                            error=True,
                                            error_log="Parsing error : Invalid end tag detected",
                                            parsed_outputs={
                                                active_key: buffer.split(start_token)[0]
                                            },
                                        )
                                        buffer = buffer.split(end_token + end_token)[-1]
                                        stream_pause = False
                                        break
                                    else:
                                        if (
                                            buffer.find(start_token + start_token) != -1
                                        ):  # if [[ in buffer -> pause
                                            break
                                        else:
                                            # if [ in buffer (== [blah]) -> stream
                                            yield LLMStreamResponse(
                                                parsed_outputs={active_key: buffer}
                                            )
                                            buffer = ""
                                            stream_pause = False
                                            break
                                break
                            else:
                                # no start token, no stream_pause (not inside of tag)
                                if buffer:
                                    yield LLMStreamResponse(
                                        parsed_outputs={active_key: buffer}
                                    )
                                    buffer = ""
                                break

                if (
                    "function_call" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["function_call"] is not None
                ):
                    for key, value in chunk["choices"][0]["delta"][
                        "function_call"
                    ].items():
                        function_call[key] += value

                if chunk["choices"][0]["finish_reason"] != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            function_list=functions,
                            function_call=function_call
                            if chunk["choices"][0]["finish_reason"] == "function_call"
                            else None,
                        ),
                        function_call=function_call
                        if chunk["choices"][0]["finish_reason"] == "function_call"
                        else None,
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    def __llm_stream_response_generator__(
        self,
        messages: List[Dict[str, str]],
        response: Generator,
        start_time: datetime.datetime,
        functions: List[Any] = [],
    ):
        raw_output = ""
        function_call = {"name": "", "arguments": ""}
        try:
            for chunk in response:
                if (
                    "content" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["content"] is not None
                ):
                    raw_output += chunk["choices"][0]["delta"]["content"]
                    yield LLMStreamResponse(
                        raw_output=chunk["choices"][0]["delta"]["content"]
                    )

                if (
                    "function_call" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["function_call"] is not None
                ):
                    for key, value in chunk["choices"][0]["delta"][
                        "function_call"
                    ].items():
                        function_call[key] += value

                if chunk["choices"][0]["finish_reason"] != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            function_list=functions,
                            function_call=function_call
                            if chunk["choices"][0]["finish_reason"] == "function_call"
                            else None,
                        ),
                        function_call=function_call
                        if chunk["choices"][0]["finish_reason"] == "function_call"
                        else None,
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    def __single_type_sp_generator__(
        self,
        messages: List[Dict[str, str]],
        response: Generator,
        parsing_type: ParsingType,
        start_time: datetime.datetime,
        functions: List[Any] = [],
    ):
        try:
            parsing_pattern = get_pattern_by_type(parsing_type)
            start_tag = parsing_pattern["start"]
            start_fstring = parsing_pattern["start_fstring"]
            end_fstring = parsing_pattern["end_fstring"]
            start_token = parsing_pattern["start_token"]
            end_token = parsing_pattern["end_token"]

            buffer = ""
            raw_output = ""
            active_key = None
            stream_pause = False
            end_tag = None
            function_call = {"name": "", "arguments": ""}
            for chunk in response:
                if (
                    "content" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["content"] is not None
                ):
                    stream_value: str = chunk["choices"][0]["delta"]["content"]
                    raw_output += stream_value
                    yield LLMStreamResponse(raw_output=stream_value)
                    buffer += stream_value

                    while True:
                        if active_key is None:
                            keys = re.findall(start_tag, buffer, flags=re.DOTALL)
                            if len(keys) > 1:
                                yield LLMStreamResponse(
                                    error=True,
                                    error_log="Parsing error : Nested key detected",
                                )
                                break
                            if len(keys) == 0:
                                break  # no key

                            active_key, active_type = keys[
                                0
                            ]  # Updated to unpack both key and type
                            end_tag = end_fstring.format(key=active_key)
                            # delete start tag from buffer
                            start_pattern = start_fstring.format(
                                key=active_key, type=active_type
                            )
                            buffer = buffer.split(start_pattern)[-1]

                        else:
                            if (
                                stream_value.find(start_token) != -1
                            ):  # start token appers in chunk -> pause
                                stream_pause = True
                                break
                            elif stream_pause:
                                if (
                                    buffer.find(end_tag) != -1
                                ):  # if end tag appears in buffer
                                    yield LLMStreamResponse(
                                        parsed_outputs={
                                            active_key: buffer.split(end_tag)[
                                                0
                                            ].replace(end_tag, "")
                                        }
                                    )
                                    buffer = buffer.split(end_tag)[-1]
                                    active_key = None
                                    stream_pause = False
                                elif (
                                    stream_value.find(end_token) != -1
                                ):  # if pattern ends  = ("[blah]" != end_pattern) appeared in buffer
                                    if (
                                        active_type == "List"
                                        or active_type == "Dict"
                                        and end_token.find("]") != -1
                                    ):
                                        try:
                                            buffer_dict = json.loads(buffer)
                                            stream_pause = False
                                            continue
                                        except Exception as exception:
                                            logger.error(exception)
                                            yield LLMStreamResponse(
                                                error=True,
                                                error_log="Parsing error : Invalid end tag detected",
                                                parsed_outputs={
                                                    active_key: buffer.split(
                                                        start_token
                                                    )[0]
                                                },
                                            )
                                            stream_pause = False
                                            buffer = ""
                                    yield LLMStreamResponse(
                                        error=True,
                                        error_log="Parsing error : Invalid end tag detected",
                                        parsed_outputs={active_key: buffer},
                                    )
                                    stream_pause = False
                                    buffer = ""
                                break
                            else:
                                # no start token, no stream_pause (not inside of tag)
                                if buffer:
                                    yield LLMStreamResponse(
                                        parsed_outputs={active_key: buffer}
                                    )
                                    buffer = ""
                                break

                if (
                    "function_call" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["function_call"] is not None
                ):
                    for key, value in chunk["choices"][0]["delta"][
                        "function_call"
                    ].items():
                        function_call[key] += value

                if chunk["choices"][0]["finish_reason"] != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            function_list=functions,
                            function_call=function_call
                            if chunk["choices"][0]["finish_reason"] == "function_call"
                            else None,
                        ),
                        function_call=function_call
                        if chunk["choices"][0]["finish_reason"] == "function_call"
                        else None,
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    async def __double_type_sp_agenerator__(
        self,
        messages: List[Dict[str, str]],
        response: AsyncGenerator,
        parsing_type: ParsingType,
        start_time: datetime.datetime,
        functions: List[Any] = [],
    ):
        try:
            parsing_pattern = get_pattern_by_type(parsing_type)
            start_tag = parsing_pattern["start"]
            start_fstring = parsing_pattern["start_fstring"]
            end_fstring = parsing_pattern["end_fstring"]
            start_token = parsing_pattern["start_token"]
            end_token = parsing_pattern["end_token"]

            buffer = ""
            raw_output = ""
            active_key = None
            stream_pause = False
            end_tag = None
            function_call = {"name": "", "arguments": ""}
            async for chunk in response:
                if (
                    "content" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["content"] is not None
                ):
                    stream_value: str = chunk["choices"][0]["delta"]["content"]
                    raw_output += stream_value
                    yield LLMStreamResponse(raw_output=stream_value)
                    buffer += stream_value

                    while True:
                        if active_key is None:
                            keys = re.findall(start_tag, buffer, flags=re.DOTALL)
                            if len(keys) > 1:
                                yield LLMStreamResponse(
                                    error=True,
                                    error_log="Parsing error : Nested key detected",
                                )
                                break
                            if len(keys) == 0:
                                break  # no key
                            active_key, active_type = keys[0]
                            end_tag = end_fstring.format(key=active_key)
                            # delete start tag from buffer
                            start_pattern = start_fstring.format(
                                key=active_key, type=active_type
                            )
                            buffer = buffer.split(start_pattern)[-1]

                        else:
                            if (
                                stream_value.find(start_token) != -1
                            ):  # start token appers in chunk -> pause
                                stream_pause = True
                                break
                            elif stream_pause:
                                if (
                                    buffer.find(end_tag) != -1
                                ):  # if end tag appears in buffer
                                    yield LLMStreamResponse(
                                        parsed_outputs={
                                            active_key: buffer.split(end_tag)[0]
                                        }
                                    )
                                    buffer = buffer.split(end_tag)[-1]
                                    active_key = None
                                    stream_pause = False
                                    # break
                                elif (
                                    stream_value.find(end_token) != -1
                                ):  # if ("[blah]" != end_pattern) appeared in buffer
                                    if (
                                        buffer.find(end_token + end_token) != -1
                                    ):  # if ]] in buffer -> error
                                        yield LLMStreamResponse(
                                            error=True,
                                            error_log="Parsing error : Invalid end tag detected",
                                            parsed_outputs={
                                                active_key: buffer.split(start_token)[0]
                                            },
                                        )
                                        buffer = buffer.split(end_token + end_token)[-1]
                                        stream_pause = False
                                        break
                                    else:
                                        if (
                                            buffer.find(start_token + start_token) != -1
                                        ):  # if [[ in buffer -> pause
                                            break
                                        else:
                                            # if [ in buffer (== [blah]) -> stream
                                            yield LLMStreamResponse(
                                                parsed_outputs={active_key: buffer}
                                            )
                                            buffer = ""
                                            stream_pause = False
                                            break
                                break
                            else:
                                # no start token, no stream_pause (not inside of tag)
                                if buffer:
                                    yield LLMStreamResponse(
                                        parsed_outputs={active_key: buffer}
                                    )
                                    buffer = ""
                                break

                if (
                    "function_call" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["function_call"] is not None
                ):
                    for key, value in chunk["choices"][0]["delta"][
                        "function_call"
                    ].items():
                        function_call[key] += value

                if chunk["choices"][0]["finish_reason"] != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            function_list=functions,
                            function_call=function_call
                            if chunk["choices"][0]["finish_reason"] == "function_call"
                            else None,
                        ),
                        function_call=function_call
                        if chunk["choices"][0]["finish_reason"] == "function_call"
                        else None,
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    async def __llm_stream_response_agenerator__(
        self,
        messages: List[Dict[str, str]],
        response: AsyncGenerator,
        start_time: datetime.datetime,
        functions: List[Any] = [],
    ):
        raw_output = ""
        function_call = {"name": "", "arguments": ""}
        try:
            async for chunk in response:
                if (
                    "content" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["content"] is not None
                ):
                    raw_output += chunk["choices"][0]["delta"]["content"]
                    yield LLMStreamResponse(
                        raw_output=chunk["choices"][0]["delta"]["content"]
                    )

                if (
                    "function_call" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["function_call"] is not None
                ):
                    for key, value in chunk["choices"][0]["delta"][
                        "function_call"
                    ].items():
                        function_call[key] += value

                if chunk["choices"][0]["finish_reason"] != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            function_list=functions,
                            function_call=function_call
                            if chunk["choices"][0]["finish_reason"] == "function_call"
                            else None,
                        ),
                        function_call=function_call
                        if chunk["choices"][0]["finish_reason"] == "function_call"
                        else None,
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    async def __single_type_sp_agenerator__(
        self,
        messages: List[Dict[str, str]],
        response: AsyncGenerator,
        parsing_type: ParsingType,
        start_time: datetime.datetime,
        functions: List[Any] = [],
    ):
        try:
            parsing_pattern = get_pattern_by_type(parsing_type)
            start_tag = parsing_pattern["start"]
            start_fstring = parsing_pattern["start_fstring"]
            end_fstring = parsing_pattern["end_fstring"]
            start_token = parsing_pattern["start_token"]
            end_token = parsing_pattern["end_token"]

            buffer = ""
            raw_output = ""
            active_key = None
            stream_pause = False
            end_tag = None
            function_call = {"name": "", "arguments": ""}
            async for chunk in response:
                if (
                    "content" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["content"] is not None
                ):
                    stream_value: str = chunk["choices"][0]["delta"]["content"]
                    raw_output += stream_value
                    yield LLMStreamResponse(raw_output=stream_value)
                    buffer += stream_value

                    while True:
                        if active_key is None:
                            keys = re.findall(start_tag, buffer, flags=re.DOTALL)
                            if len(keys) > 1:
                                yield LLMStreamResponse(
                                    error=True,
                                    error_log="Parsing error : Nested key detected",
                                )
                                break
                            if len(keys) == 0:
                                break  # no key

                            active_key, active_type = keys[
                                0
                            ]  # Updated to unpack both key and type
                            end_tag = end_fstring.format(key=active_key)
                            # delete start tag from buffer
                            start_pattern = start_fstring.format(
                                key=active_key, type=active_type
                            )
                            buffer = buffer.split(start_pattern)[-1]

                        else:
                            if (
                                stream_value.find(start_token) != -1
                            ):  # start token appers in chunk -> pause
                                stream_pause = True
                                break
                            elif stream_pause:
                                if (
                                    buffer.find(end_tag) != -1
                                ):  # if end tag appears in buffer
                                    yield LLMStreamResponse(
                                        parsed_outputs={
                                            active_key: buffer.split(end_tag)[
                                                0
                                            ].replace(end_tag, "")
                                        }
                                    )
                                    buffer = buffer.split(end_tag)[-1]
                                    active_key = None
                                    stream_pause = False
                                elif (
                                    stream_value.find(end_token) != -1
                                ):  # if pattern ends  = ("[blah]" != end_pattern) appeared in buffer
                                    if (
                                        active_type == "List"
                                        or active_type == "Dict"
                                        and end_token.find("]") != -1
                                    ):
                                        try:
                                            buffer_dict = json.loads(buffer)
                                            stream_pause = False
                                            continue
                                        except Exception as exception:
                                            logger.error(exception)
                                            yield LLMStreamResponse(
                                                error=True,
                                                error_log="Parsing error : Invalid end tag detected",
                                                parsed_outputs={
                                                    active_key: buffer.split(
                                                        start_token
                                                    )[0]
                                                },
                                            )
                                            stream_pause = False
                                            buffer = ""
                                    yield LLMStreamResponse(
                                        error=True,
                                        error_log="Parsing error : Invalid end tag detected",
                                        parsed_outputs={active_key: buffer},
                                    )
                                    stream_pause = False
                                    buffer = ""
                                break
                            else:
                                # no start token, no stream_pause (not inside of tag)
                                if buffer:
                                    yield LLMStreamResponse(
                                        parsed_outputs={active_key: buffer}
                                    )
                                    buffer = ""
                                break

                if (
                    "function_call" in chunk["choices"][0]["delta"]
                    and chunk["choices"][0]["delta"]["function_call"] is not None
                ):
                    for key, value in chunk["choices"][0]["delta"][
                        "function_call"
                    ].items():
                        function_call[key] += value

                if chunk["choices"][0]["finish_reason"] != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            function_list=functions,
                            function_call=function_call
                            if chunk["choices"][0]["finish_reason"] == "function_call"
                            else None,
                        ),
                        function_call=function_call
                        if chunk["choices"][0]["finish_reason"] == "function_call"
                        else None,
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))



import asyncio
import json
import datetime
import re

from uuid import UUID
from typing import Dict, Any, Optional, AsyncGenerator, List
from dotenv import load_dotenv

from websockets.client import connect, WebSocketClientProtocol
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
from readerwriterlock import rwlock
from playhouse.shortcuts import model_to_dict

import promptmodel.utils.logger as logger
from promptmodel import DevApp
from promptmodel.llms.llm_dev import LLMDev
from promptmodel.database.crud import (
    create_prompt_model_version,
    create_prompt,
    create_run_log,
    list_prompt_models,
    list_prompt_model_versions,
    list_prompts,
    list_run_logs,
    list_samples,
    get_sample_input,
    get_prompt_model_uuid,
    update_prompt_model_version,
    find_ancestor_version,
    find_ancestor_versions,
    update_candidate_prompt_model_version,
)
from promptmodel.database.models import PromptModelVersion, PromptModel
from promptmodel.utils.enums import (
    ServerTask,
    LocalTask,
    PromptModelVersionStatus,
)
from promptmodel.utils.config_utils import upsert_config, read_config
from promptmodel.utils.types import LLMStreamResponse
from promptmodel.constants import ENDPOINT_URL

load_dotenv()
GATEWAY_URL = f"wss://{ENDPOINT_URL.split('://')[1]}/open_websocket"


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            aware_datetime = obj.replace(tzinfo=datetime.timezone.utc)
            return aware_datetime.isoformat()  # This will include timezone information
        return super().default(obj)


class DevWebsocketClient:
    def __init__(self, _devapp: DevApp):
        self._devapp: DevApp = _devapp
        self.rwlock = rwlock.RWLockFair()

    async def _get_prompt_models(self, prompt_model_name: str):
        """Get prompt_model from registry"""
        with self.rwlock.gen_rlock():
            prompt_model = next(
                (
                    prompt_model
                    for prompt_model in self._devapp.prompt_models
                    if prompt_model.name == prompt_model_name
                ),
                None,
            )
        return prompt_model

    def update_devapp_instance(self, new_devapp):
        with self.rwlock.gen_wlock():
            self._devapp = new_devapp
            if self.ws:
                asyncio.run(
                    self.ws.send(
                        json.dumps({"type": ServerTask.LOCAL_UPDATE_ALERT.value})
                    )
                )

    async def __handle_message(
        self, message: Dict[str, Any], ws: WebSocketClientProtocol
    ):
        # logger.info(f"Received message: {message}")
        response: Dict[Any, str] = {}
        # If the message has a correlation_id, add it to the response
        # correlation_id is the unique ID of the function from backend to local
        if message.get("correlation_id"):
            response["correlation_id"] = message["correlation_id"]

        # If the message has a runner_id, add it to the response
        if message.get("runner_id"):
            response["runner_id"] = message["runner_id"]

        data = None
        try:
            if message["type"] == LocalTask.LIST_PROMPT_MODELS:
                res_from_local_db = list_prompt_models()
                modules_with_used_in_code = [
                    module
                    for module in res_from_local_db
                    if module["used_in_code"] == True
                ]
                data = {"prompt_models": modules_with_used_in_code}

            elif message["type"] == LocalTask.LIST_PROMPT_MODEL_VERSIONS:
                prompt_model_uuid = message["prompt_model_uuid"]
                res_from_local_db = list_prompt_model_versions(prompt_model_uuid)
                data = {"prompt_model_versions": res_from_local_db}

            elif message["type"] == LocalTask.LIST_SAMPLES:
                res_from_local_db = list_samples()
                data = {"samples": res_from_local_db}

            elif message["type"] == LocalTask.LIST_FUNCTIONS:
                function_name_list = self._devapp._get_function_name_list()
                data = {"functions": function_name_list}

            elif message["type"] == LocalTask.GET_PROMPTS:
                prompt_model_version_uuid = message["prompt_model_version_uuid"]
                res_from_local_db = list_prompts(prompt_model_version_uuid)
                data = {"prompts": res_from_local_db}

            elif message["type"] == LocalTask.GET_RUN_LOGS:
                prompt_model_version_uuid = message["prompt_model_version_uuid"]
                res_from_local_db = list_run_logs(prompt_model_version_uuid)
                data = {"run_logs": res_from_local_db}

            elif message["type"] == LocalTask.CHANGE_PROMPT_MODEL_VERSION_STATUS:
                prompt_model_version_uuid = message["prompt_model_version_uuid"]
                new_status = message["status"]
                res_from_local_db = update_prompt_model_version(
                    prompt_model_version_uuid=prompt_model_version_uuid,
                    status=new_status,
                )
                data = {
                    "prompt_model_version_uuid": prompt_model_version_uuid,
                    "status": new_status,
                }

            elif message["type"] == LocalTask.GET_PROMPT_MODEL_VERSION_TO_SAVE:
                prompt_model_version_uuid = message["prompt_model_version_uuid"]
                # change from_uuid to candidate ancestor
                prompt_model_version, prompts = find_ancestor_version(
                    prompt_model_version_uuid
                )
                # delete status, is_published
                del prompt_model_version["status"]
                del prompt_model_version["is_published"]
                config = read_config()
                prompt_model_version["dev_branch_uuid"] = config["dev_branch"]["uuid"]

                for prompt in prompts:
                    del prompt["id"]

                prompt_model = PromptModel.get(
                    PromptModel.uuid == prompt_model_version.prompt_model_uuid
                ).__data__
                is_deployed = prompt_model["is_deployed"]
                if not is_deployed:
                    data = {
                        "prompt_model": {
                            "uuid": prompt_model["uuid"],
                            "name": prompt_model["name"],
                            "project_uuid": prompt_model["project_uuid"],
                        },
                        "version": prompt_model_version,
                        "prompts": prompts,
                    }
                else:
                    data = {
                        "prompt_model": None,
                        "version": prompt_model_version,
                        "prompts": prompts,
                    }

            elif message["type"] == LocalTask.GET_PROMPT_MODEL_VERSIONS_TO_SAVE:
                target_prompt_model_uuid = (
                    message["prompt_model_uuid"]
                    if "prompt_model_uuid" in message
                    else None
                )

                prompt_model_versions, prompts = find_ancestor_versions(
                    target_prompt_model_uuid
                )

                config = read_config()
                for prompt_model_version in prompt_model_versions:
                    prompt_model_version["dev_branch_uuid"] = config["dev_branch"][
                        "uuid"
                    ]
                    del prompt_model_version["id"]
                    del prompt_model_version["status"]
                    del prompt_model_version["is_published"]

                for prompt in prompts:
                    del prompt["id"]

                prompt_model_uuids = [
                    version["prompt_model_uuid"] for version in prompt_model_versions
                ]
                prompt_models = list(
                    PromptModel.select().where(PromptModel.uuid.in_(prompt_model_uuids))
                )
                prompt_models = [
                    model_to_dict(prompt_model) for prompt_model in prompt_models
                ]
                # find prompt_model which is not deployed
                prompt_models_only_in_local = []
                for prompt_model in prompt_models:
                    if prompt_model["is_deployed"] is False:
                        del prompt_model["is_deployed"]
                        del prompt_model["used_in_code"]
                        del prompt_model["id"]

                        prompt_models_only_in_local.append(prompt_model)
                data = {
                    "prompt_models": prompt_models_only_in_local,
                    "versions": prompt_model_versions,
                    "prompts": prompts,
                }

            elif message["type"] == LocalTask.UPDATE_CANDIDATE_PROMPT_MODEL_VERSION_ID:
                new_candidates = message["new_candidates"]
                update_candidate_prompt_model_version(new_candidates)

            elif message["type"] == LocalTask.RUN_PROMPT_MODEL:
                prompt_model_name: str = message["prompt_model_name"]
                sample_name: Optional[str] = message["sample_name"]

                # get sample from db
                if sample_name:
                    sample_input_row = get_sample_input(sample_name)
                    if sample_input_row is None or "contents" not in sample_input_row:
                        logger.error(f"There is no sample input {sample_name}.")
                        return
                    sample_input = sample_input_row["contents"]
                else:
                    sample_input = None

                # Check prompt_model in Local Usage
                prompt_model_names = self._devapp._get_prompt_model_name_list()
                if prompt_model_name not in prompt_model_names:
                    logger.error(f"There is no prompt_model {prompt_model_name}.")
                    return

                # Validate Variable Matching
                prompt_variables = []
                for prompt in message["prompts"]:
                    fstring_input_pattern = r"(?<!\\)(?<!{{)\{([^}]+)\}(?!\\)(?!}})"
                    prompt_variables += re.findall(
                        fstring_input_pattern, prompt["content"]
                    )
                prompt_variables = list(set(prompt_variables))
                if len(prompt_variables) != 0:
                    if sample_input is None:
                        data = {
                            "type": ServerTask.UPDATE_RESULT_RUN.value,
                            "status": "failed",
                            "log": f"Prompts have variables {prompt_variables}. You should select sample input.",
                        }
                        data.update(response)
                        # logger.debug(f"Sent response: {data}")
                        await ws.send(json.dumps(data, cls=CustomJSONEncoder))
                        return

                    if not all(
                        variable in sample_input.keys() for variable in prompt_variables
                    ):
                        missing_variables = [
                            variable
                            for variable in prompt_variables
                            if variable not in sample_input.keys()
                        ]
                        data = {
                            "type": ServerTask.UPDATE_RESULT_RUN.value,
                            "status": "failed",
                            "log": f"Sample input does not have variables {missing_variables} in prompts.",
                        }
                        data.update(response)
                        # logger.debug(f"Sent response: {data}")
                        await ws.send(json.dumps(data, cls=CustomJSONEncoder))
                        return

                # Start PromptModel Running
                output = {"raw_output": "", "parsed_outputs": {}}
                try:
                    logger.info(f"Started PromptModel: {prompt_model_name}")
                    # create prompt_model_dev_instance
                    prompt_model_dev = LLMDev()
                    # fine prompt_model_uuid from local db
                    prompt_model_uuid: str = get_prompt_model_uuid(prompt_model_name)[
                        "uuid"
                    ]

                    prompt_model_version_uuid: Optional[str] = message["uuid"]
                    # If prompt_model_version_uuid is None, create new version & prompt
                    if prompt_model_version_uuid is None:
                        prompt_model_version: PromptModelVersion = (
                            create_prompt_model_version(
                                prompt_model_uuid=prompt_model_uuid,
                                status=PromptModelVersionStatus.BROKEN.value,
                                from_uuid=message["from_uuid"],
                                model=message["model"],
                                parsing_type=message["parsing_type"],
                                output_keys=message["output_keys"],
                                functions=message["functions"],
                            )
                        )
                        prompt_model_version_uuid: str = prompt_model_version.uuid

                        prompts = message["prompts"]
                        for prompt in prompts:
                            create_prompt(
                                version_uuid=prompt_model_version_uuid,
                                role=prompt["role"],
                                step=prompt["step"],
                                content=prompt["content"],
                            )
                        # send message to backend
                        data = {
                            "type": ServerTask.UPDATE_RESULT_RUN.value,
                            "prompt_model_version_uuid": prompt_model_version_uuid,
                            "status": "running",
                        }
                        data.update(response)
                        # logger.debug(f"Sent response: {data}")
                        await ws.send(json.dumps(data, cls=CustomJSONEncoder))

                    data = {
                        "type": ServerTask.UPDATE_RESULT_RUN.value,
                        "status": "running",
                        "inputs": sample_input if sample_input else {},
                    }
                    data.update(response)
                    # logger.debug(f"Sent response: {data}")
                    await ws.send(json.dumps(data, cls=CustomJSONEncoder))

                    model = message["model"]
                    prompts = message["prompts"]
                    parsing_type = message["parsing_type"]

                    if sample_input:
                        messages_for_run = [
                            {
                                "content": prompt["content"].format(**sample_input),
                                "role": prompt["role"],
                            }
                            for prompt in prompts
                        ]
                    else:
                        messages_for_run = prompts

                    parsing_success = True
                    error_log = None
                    function_call = None
                    function_call_log = None

                    # get function schemas from register & send to LLM
                    function_names: List[str] = message["functions"]
                    logger.debug(f"function_names : {function_names}")
                    try:
                        function_schemas: List[
                            str
                        ] = self._devapp._get_function_schemas(function_names)
                    except Exception as error:
                        data = {
                            "type": ServerTask.UPDATE_RESULT_RUN.value,
                            "status": "failed",
                            "log": f"There is no function. : {error}",
                        }
                        data.update(response)
                        # logger.debug(f"Sent response: {data}")
                        await ws.send(json.dumps(data, cls=CustomJSONEncoder))
                        return

                    logger.debug(f"functions : {function_schemas}")

                    res: AsyncGenerator[
                        LLMStreamResponse, None
                    ] = prompt_model_dev.dev_run(
                        messages=messages_for_run,
                        parsing_type=parsing_type,
                        functions=function_schemas,
                        model=model,
                    )
                    async for item in res:
                        # send item to backend
                        # save item & parse
                        # if type(item) == str: raw output, if type(item) == dict: parsed output
                        if item.raw_output is not None:
                            output["raw_output"] += item.raw_output
                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "running",
                                "raw_output": item.raw_output,
                            }
                        if item.parsed_outputs:
                            if (
                                list(item.parsed_outputs.keys())[0]
                                not in output["parsed_outputs"]
                            ):
                                output["parsed_outputs"][
                                    list(item.parsed_outputs.keys())[0]
                                ] = list(item.parsed_outputs.values())[0]
                            else:
                                output["parsed_outputs"][
                                    list(item.parsed_outputs.keys())[0]
                                ] += list(item.parsed_outputs.values())[0]
                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "running",
                                "parsed_outputs": item.parsed_outputs,
                            }
                        if item.function_call is not None:
                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "running",
                                "function_call": item.function_call,
                            }
                            function_call = item.function_call

                        if item.error and parsing_success is True:
                            parsing_success = not item.error
                            error_log = item.error_log

                        data.update(response)
                        # logger.debug(f"Sent response: {data}")
                        await ws.send(json.dumps(data, cls=CustomJSONEncoder))

                    # IF function_call in response -> call function -> call LLM once more
                    if function_call:
                        # make function_call_log
                        function_call_log = {
                            "name": function_call["name"],
                            "arguments": function_call["arguments"],
                            "response": None,
                            "initial_raw_output": output["raw_output"],
                        }

                        # call function
                        try:
                            function_call_args: Dict[str, Any] = json.loads(
                                function_call["arguments"]
                            )
                            function_response = self._devapp._call_register_function(
                                function_call["name"], function_call_args
                            )
                            function_call_log["response"] = function_response
                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "running",
                                "function_response": function_response,
                            }
                            data.update(response)
                            # logger.debug(f"Sent response: {data}")
                            await ws.send(json.dumps(data, cls=CustomJSONEncoder))
                        except Exception as error:
                            logger.error(f"{error}")

                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "failed",
                                "log": f"Function call Failed, {error}",
                            }
                            update_prompt_model_version(
                                prompt_model_version_uuid=prompt_model_version_uuid,
                                status=PromptModelVersionStatus.BROKEN.value,
                            )
                            create_run_log(
                                prompt_model_version_uuid=prompt_model_version_uuid,
                                inputs=sample_input,
                                raw_output=output["raw_output"],
                                parsed_outputs=output["parsed_outputs"],
                                function_call=function_call_log,
                            )
                            response.update(data)
                            await ws.send(json.dumps(response, cls=CustomJSONEncoder))
                            return
                        # call LLM once more
                        messages_for_run += [
                            {
                                "role": "assistant",
                                "function_call": function_call,
                            },
                            {
                                "role": "function",
                                "name": function_call["name"],
                                "content": str(function_response),
                            },
                        ]
                        res_after_function_call: AsyncGenerator[
                            LLMStreamResponse, None
                        ] = prompt_model_dev.dev_chat(
                            messages_for_run, parsing_type, model
                        )

                        output = {"raw_output": "", "parsed_outputs": {}}
                        async for item in res_after_function_call:
                            if item.raw_output is not None:
                                output["raw_output"] += item.raw_output
                                data = {
                                    "type": ServerTask.UPDATE_RESULT_RUN.value,
                                    "status": "running",
                                    "raw_output": item.raw_output,
                                }
                            if item.parsed_outputs:
                                if (
                                    list(item.parsed_outputs.keys())[0]
                                    not in output["parsed_outputs"]
                                ):
                                    output["parsed_outputs"][
                                        list(item.parsed_outputs.keys())[0]
                                    ] = list(item.parsed_outputs.values())[0]
                                else:
                                    output["parsed_outputs"][
                                        list(item.parsed_outputs.keys())[0]
                                    ] += list(item.parsed_outputs.values())[0]
                                data = {
                                    "type": ServerTask.UPDATE_RESULT_RUN.value,
                                    "status": "running",
                                    "parsed_outputs": item.parsed_outputs,
                                }

                            if item.error and parsing_success is True:
                                parsing_success = not item.error
                                error_log = item.error_log

                            data.update(response)
                            # logger.debug(f"Sent response: {data}")
                            await ws.send(json.dumps(data, cls=CustomJSONEncoder))

                    if (
                        message["output_keys"] is not None
                        and message["parsing_type"] is not None
                        and set(output["parsed_outputs"].keys())
                        != set(message["output_keys"])
                        or (parsing_success is False)
                    ):
                        error_log = error_log if error_log else "Key matching failed."
                        data = {
                            "type": ServerTask.UPDATE_RESULT_RUN.value,
                            "status": "failed",
                            "log": f"parsing failed, {error_log}",
                        }
                        update_prompt_model_version(
                            prompt_model_version_uuid=prompt_model_version_uuid,
                            status=PromptModelVersionStatus.BROKEN.value,
                        )
                        create_run_log(
                            prompt_model_version_uuid=prompt_model_version_uuid,
                            inputs=sample_input,
                            raw_output=output["raw_output"],
                            parsed_outputs=output["parsed_outputs"],
                            function_call=function_call_log,
                        )
                        response.update(data)
                        await ws.send(json.dumps(response, cls=CustomJSONEncoder))
                        return

                    data = {
                        "type": ServerTask.UPDATE_RESULT_RUN.value,
                        "status": "completed",
                    }
                    update_prompt_model_version(
                        prompt_model_version_uuid=prompt_model_version_uuid,
                        status=PromptModelVersionStatus.WORKING.value,
                    )

                    create_run_log(
                        prompt_model_version_uuid=prompt_model_version_uuid,
                        inputs=sample_input,
                        raw_output=output["raw_output"],
                        parsed_outputs=output["parsed_outputs"],
                        function_call=function_call_log,
                    )
                except Exception as error:
                    logger.error(f"Error running service: {error}")
                    data = {
                        "type": ServerTask.UPDATE_RESULT_RUN.value,
                        "status": "failed",
                        "log": str(error),
                    }
                    response.update(data)
                    await ws.send(json.dumps(response, cls=CustomJSONEncoder))
                    return

            if data:
                response.update(data)
                await ws.send(json.dumps(response, cls=CustomJSONEncoder))
            logger.info(f"Sent response: {response}")
        except Exception as error:
            logger.error(f"Error handling message: {error}")
            await ws.send(str(error))

    async def connect_to_gateway(
        self,
        project_uuid: str,
        dev_branch_name: str,
        cli_access_header: dict,
        retries=12 * 24,
        retry_delay=5 * 60,
    ):
        """Open Websocket to Backend with project_uuid, dev_branch_name, cli_access_token"""
        headers = cli_access_header
        headers.update(
            {"project_uuid": project_uuid, "dev_branch_name": dev_branch_name}
        )
        for _ in range(retries):
            try:
                async with connect(
                    GATEWAY_URL,
                    extra_headers=headers,
                    # ping_interval=10,
                    # ping_timeout=1,
                    # timeout=3600 * 24,  # Timeout is set to 24 hours
                ) as ws:
                    logger.success("Connected to gateway. Your DevApp is now online! 🎉")
                    self.ws = ws
                    while True:
                        message = await ws.recv()
                        data = json.loads(message)
                        await self.__handle_message(data, ws)
            except (ConnectionClosedError, ConnectionClosedOK):
                # If the connection was closed gracefully, handle it accordingly
                logger.warning("Connection to the gateway was closed.")
            except TimeoutError:
                logger.error(
                    f"Timeout error while connecting to the gateway. Retrying in {retry_delay} seconds..."
                )
                await asyncio.sleep(retry_delay)
            except Exception as error:
                logger.error(f"Error receiving message: {error}")



import os
import sys
import importlib
from typing import Any, Dict, List
from threading import Timer
from rich import print
from watchdog.events import FileSystemEventHandler

from promptmodel.apis.base import APIClient
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.utils import logger
from promptmodel import DevApp, DevClient
from promptmodel.database.crud import (
    list_prompt_model_versions,
    list_prompt_models,
    update_used_in_code_prompt_model_by_name,
    create_prompt_models,
    create_prompt_model_versions,
    create_prompts,
    create_run_logs,
    update_samples,
    get_prompt_model_uuid,
    update_prompt_model_uuid,
)
from promptmodel.utils.enums import (
    PromptModelVersionStatus,
    ChangeLogAction,
)
from promptmodel.websocket.websocket_client import DevWebsocketClient


class CodeReloadHandler(FileSystemEventHandler):
    def __init__(
        self,
        _devapp_filename: str,
        _instance_name: str,
        dev_websocket_client: DevWebsocketClient,
    ):
        self._devapp_filename: str = _devapp_filename
        self.devapp_instance_name: str = _instance_name
        self.dev_websocket_client: DevWebsocketClient = (
            dev_websocket_client  # save dev_websocket_client instance
        )
        self.timer = None

    def on_modified(self, event):
        """Called when a file or directory is modified."""
        if event.src_path.endswith(".py"):
            if self.timer:
                self.timer.cancel()
            # reload modified file & main file
            self.timer = Timer(0.5, self.reload_code, args=(event.src_path,))
            self.timer.start()

    def reload_code(self, modified_file_path: str):
        print(
            f"[violet]promptmodel:dev:[/violet]  Reloading {self._devapp_filename} module due to changes..."
        )
        relative_modified_path = os.path.relpath(modified_file_path, os.getcwd())
        # Reload the devapp module
        module_name = relative_modified_path.replace("./", "").replace("/", ".")[
            :-3
        ]  # assuming the file is in the PYTHONPATH

        if module_name in sys.modules:
            module = sys.modules[module_name]
            importlib.reload(module)

        reloaded_module = importlib.reload(sys.modules[self._devapp_filename])
        print(
            f"[violet]promptmodel:dev:[/violet]  {self._devapp_filename} module reloaded successfully."
        )

        new_devapp_instance: DevApp = getattr(
            reloaded_module, self.devapp_instance_name
        )
        # print(new_devapp_instance.prompt_models)
        new_prompt_model_name_list = [
            prompt_model.name for prompt_model in new_devapp_instance.prompt_models
        ]
        old_prompt_model_name_list = [
            prompt_model.name
            for prompt_model in self.dev_websocket_client._devapp.prompt_models
        ]

        # 사라진 prompt_models 에 대해 local db prompt_model.used_in_code False Update
        removed_name_list = list(
            set(old_prompt_model_name_list) - set(new_prompt_model_name_list)
        )
        for removed_name in removed_name_list:
            update_used_in_code_prompt_model_by_name(removed_name, False)

        # 새로 생긴 prompt_model 에 대해 local db prompt_model.used_in_code True Update
        # TODO: 좀 더 specific 한 API와 연결 필요
        config = read_config()
        org = config["dev_branch"]["org"]
        project = config["dev_branch"]["project"]
        project_status = APIClient.execute(
            method="GET",
            path="/pull_project",
            params={"project_uuid": project["uuid"]},
        ).json()

        changelogs = APIClient.execute(
            method="GET",
            path="/get_changelog",
            params={
                "project_uuid": project["uuid"],
                "local_project_version": config["dev_branch"]["project_version"],
                "levels": [1, 2],
            },
        ).json()
        # IF used_in_code=False 인 name=name 이 있을 경우, used_in_code=True
        update_by_changelog_for_reload(
            changelogs=changelogs,
            project_status=project_status,
            local_code_prompt_model_name_list=new_prompt_model_name_list,
        )

        for prompt_model in new_devapp_instance.prompt_models:
            if prompt_model.name not in old_prompt_model_name_list:
                update_used_in_code_prompt_model_by_name(prompt_model.name, True)

        # create prompt_models in local DB
        db_prompt_model_list = list_prompt_models()
        db_prompt_model_name_list = [x["name"] for x in db_prompt_model_list]
        only_in_local_names = list(
            set(new_prompt_model_name_list) - set(db_prompt_model_name_list)
        )
        only_in_local_prompt_models = [
            {"name": x, "project_uuid": project["uuid"]} for x in only_in_local_names
        ]
        create_prompt_models(only_in_local_prompt_models)

        # update samples in local DB
        update_samples(new_devapp_instance.samples)
        self.dev_websocket_client.update_devapp_instance(new_devapp_instance)


def update_by_changelog_for_reload(
    changelogs: List[Dict],
    project_status: dict,
    local_code_prompt_model_name_list: List[str],
):
    """Update Local DB by changelog"""
    local_db_prompt_model_list: list = list_prompt_models()  # {"name", "uuid"}

    for changelog in changelogs:
        level: int = changelog["level"]
        logs = changelog["logs"]
        if level == 1:
            for log in logs:
                subject = log["subject"]
                action: str = log["action"]
                if subject == "prompt_model":
                    local_db_prompt_model_list = update_prompt_model_changelog(
                        action=action,
                        project_status=project_status,
                        uuid_list=log["identifiers"],
                        local_db_prompt_model_list=local_db_prompt_model_list,
                        local_code_prompt_model_name_list=local_code_prompt_model_name_list,
                    )
                elif subject == "prompt_model_version":
                    local_db_prompt_model_list = update_prompt_model_version_changelog(
                        action=action,
                        project_status=project_status,
                        uuid_list=log["identifiers"],
                        local_db_prompt_model_list=local_db_prompt_model_list,
                        local_code_prompt_model_name_list=local_code_prompt_model_name_list,
                    )
                else:
                    pass
            previous_version_levels = changelog["previous_version"].split(".")
            current_version_levels = [
                str(int(previous_version_levels[0]) + 1),
                "0",
                "0",
            ]
            current_version = ".".join(current_version_levels)
        elif level == 2:
            for log in logs:
                subject = log["subject"]
                action: str = log["action"]
                uuid_list: list = log["identifiers"]
                if subject == "prompt_model_version":
                    local_db_prompt_model_list = update_prompt_model_version_changelog(
                        action=action,
                        project_status=project_status,
                        uuid_list=log["identifiers"],
                        local_db_prompt_model_list=local_db_prompt_model_list,
                        local_code_prompt_model_name_list=local_code_prompt_model_name_list,
                    )
                else:
                    pass
            previous_version_levels = changelog["previous_version"].split(".")
            current_version_levels = [
                previous_version_levels[0],
                str(int(previous_version_levels[1]) + 1),
                "0",
            ]
            current_version = ".".join(current_version_levels)
        else:
            previous_version_levels = changelog["previous_version"].split(".")
            current_version_levels = [
                previous_version_levels[0],
                previous_version_levels[1],
                str(int(previous_version_levels[2]) + 1),
            ]
            current_version = ".".join(current_version_levels)

        upsert_config({"project_version": current_version}, section="dev_branch")
    return True


def update_prompt_model_changelog(
    action: ChangeLogAction,
    project_status: dict,
    uuid_list: List[str],
    local_db_prompt_model_list: List[Dict],
    local_code_prompt_model_name_list: List[str],
):
    if action == ChangeLogAction.ADD.value:
        prompt_model_list = [
            x for x in project_status["prompt_models"] if x["uuid"] in uuid_list
        ]
        for prompt_model in prompt_model_list:
            local_db_prompt_model_name_list = [
                x["name"] for x in local_db_prompt_model_list
            ]

            if prompt_model["name"] not in local_db_prompt_model_name_list:
                # IF prompt_model not in Local DB
                if prompt_model["name"] in local_code_prompt_model_name_list:
                    # IF prompt_model in Local Code
                    prompt_model["used_in_code"] = True
                    prompt_model["is_deployed"] = True
                    create_prompt_models([prompt_model])
                else:
                    prompt_model["used_in_code"] = False
                    prompt_model["is_deployed"] = True
                    create_prompt_models([prompt_model])
            else:
                # Fix UUID of prompt_model
                local_uuid = get_prompt_model_uuid(prompt_model["name"])["uuid"]
                update_prompt_model_uuid(local_uuid, prompt_model["uuid"])
                local_db_prompt_model_list: list = list_prompt_models()
    else:
        # TODO: add code DELETE, CHANGE, FIX later
        pass

    return local_db_prompt_model_list


def update_prompt_model_version_changelog(
    action: ChangeLogAction,
    project_status: dict,
    uuid_list: List[str],
    local_db_prompt_model_list: List[Dict],
    local_code_prompt_model_name_list: List[str],
) -> List[Dict[str, Any]]:
    local_db_prompt_model_version_list: List[Dict] = []
    for uuid in local_db_prompt_model_list:
        local_db_prompt_model_version_list += list_prompt_model_versions(uuid["uuid"])
    uuid_list = list(
        filter(
            lambda uuid: uuid
            not in [str(x["uuid"]) for x in local_db_prompt_model_version_list],
            uuid_list,
        )
    )
    if action == ChangeLogAction.ADD.value:
        # find prompt_model_version in project_status['prompt_model_versions'] where uuid in uuid_list
        prompt_model_version_list_to_update = [
            x for x in project_status["prompt_model_versions"] if x["uuid"] in uuid_list
        ]
        # check if prompt_model_version['name'] is in local_code_prompt_model_list

        # find prompts and run_logs to update
        prompts_to_update = [
            x for x in project_status["prompts"] if x["version_uuid"] in uuid_list
        ]
        run_logs_to_update = [
            x for x in project_status["run_logs"] if x["version_uuid"] in uuid_list
        ]

        for prompt_model_version in prompt_model_version_list_to_update:
            prompt_model_version["status"] = PromptModelVersionStatus.CANDIDATE.value

        create_prompt_model_versions(prompt_model_version_list_to_update)
        create_prompts(prompts_to_update)
        create_run_logs(run_logs_to_update)

        # local_db_prompt_model_list += [{"name" : x['name'], "uuid" : x['uuid']} for x in prompt_model_version_list_to_update]
        return local_db_prompt_model_list
    else:
        pass



import asyncio
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)
from rich import print
from litellm.utils import ModelResponse
from promptmodel.llms.llm import LLM
from promptmodel.utils import logger
from promptmodel.utils.prompt_util import fetch_prompts, run_async_in_sync
from promptmodel.utils.output_utils import update_dict
from promptmodel.apis.base import APIClient, AsyncAPIClient
from promptmodel.utils.types import LLMResponse, LLMStreamResponse


class LLMProxy(LLM):
    def __init__(self, name: str, rate_limit_manager=None):
        super().__init__(rate_limit_manager)
        self._name = name

    def _wrap_gen(self, gen: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(inputs: Dict[str, Any], **kwargs):
            prompts, version_details = run_async_in_sync(fetch_prompts(self._name))
            call_args = self._prepare_call_args(
                prompts, version_details, inputs, kwargs
            )
            # Call the generator with the arguments
            stream_response: Generator[LLMStreamResponse, None, None] = gen(**call_args)

            raw_response = None
            dict_cache = {}  # to store aggregated dictionary values
            string_cache = ""  # to store aggregated string values
            error_occurs = False
            error_log = None
            for item in stream_response:
                if item.api_response:
                    raw_response = item.api_response
                if item.parsed_outputs:
                    dict_cache = update_dict(dict_cache, item.parsed_outputs)
                if item.raw_output:
                    string_cache += item.raw_output
                if item.error and not error_occurs:
                    error_occurs = True
                    error_log = item.error_log
                yield item

            # add string_cache in model_response
            if raw_response:
                if "message" not in raw_response.choices[0]:
                    raw_response.choices[0]["message"] = {}
                if "content" not in raw_response.choices[0]["message"]:
                    raw_response.choices[0]["message"]["content"] = string_cache
                    raw_response.choices[0]["message"]["role"] = "assistant"

            metadata = {
                "error_occurs": error_occurs,
                "error_log": error_log,
            }
            self._sync_log_to_cloud(
                version_details["uuid"], inputs, raw_response, dict_cache, metadata
            )

        return wrapper

    def _wrap_async_gen(self, async_gen: Callable[..., Any]) -> Callable[..., Any]:
        async def wrapper(inputs: Dict[str, Any], **kwargs):
            prompts, version_details = await fetch_prompts(self._name)
            call_args = self._prepare_call_args(
                prompts, version_details, inputs, kwargs
            )

            # Call async_gen with the arguments
            stream_response: AsyncGenerator[LLMStreamResponse, None] = async_gen(
                **call_args
            )

            raw_response = None
            dict_cache = {}  # to store aggregated dictionary values
            string_cache = ""  # to store aggregated string values
            error_occurs = False
            error_log = None
            raw_response: Optional[ModelResponse] = None
            async for item in stream_response:
                if item.api_response:
                    raw_response = item.api_response
                if item.parsed_outputs:
                    dict_cache = update_dict(dict_cache, item.parsed_outputs)
                if item.raw_output:
                    string_cache += item.raw_output
                if item.error and not error_occurs:
                    error_occurs = True
                    error_log = item.error_log
                yield item

            # add string_cache in model_response
            if raw_response:
                if "message" not in raw_response.choices[0]:
                    raw_response.choices[0]["message"] = {}
                if "content" not in raw_response.choices[0]["message"]:
                    raw_response.choices[0]["message"]["content"] = string_cache
                    raw_response.choices[0]["message"]["role"] = "assistant"

            metadata = {
                "error_occurs": error_occurs,
                "error_log": error_log,
            }
            await self._async_log_to_cloud(
                version_details["uuid"], inputs, raw_response, dict_cache, metadata
            )

            # raise Exception("error_log")

        return wrapper

    def _wrap_method(self, method: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(inputs: Dict[str, Any], **kwargs):
            prompts, version_details = run_async_in_sync(fetch_prompts(self._name))
            call_args = self._prepare_call_args(
                prompts, version_details, inputs, kwargs
            )

            # Call the method with the arguments
            llm_response: LLMResponse = method(**call_args)
            error_occurs = llm_response.error
            error_log = llm_response.error_log
            metadata = {
                "error_occurs": error_occurs,
                "error_log": error_log,
            }
            if llm_response.parsed_outputs:
                self._sync_log_to_cloud(
                    version_details["uuid"],
                    inputs,
                    llm_response.api_response,
                    llm_response.parsed_outputs,
                    metadata,
                )
            else:
                self._sync_log_to_cloud(
                    version_details["uuid"],
                    inputs,
                    llm_response.api_response,
                    {},
                    metadata,
                )
            return llm_response

        return wrapper

    def _wrap_async_method(self, method: Callable[..., Any]) -> Callable[..., Any]:
        async def async_wrapper(inputs: Dict[str, Any], **kwargs):
            prompts, version_details = await fetch_prompts(
                self._name
            )  # messages, model, uuid = self._fetch_prompts()
            call_args = self._prepare_call_args(
                prompts, version_details, inputs, kwargs
            )

            # Call the method with the arguments
            llm_response: LLMResponse = await method(**call_args)
            error_occurs = llm_response.error
            error_log = llm_response.error_log
            metadata = {
                "error_occurs": error_occurs,
                "error_log": error_log,
            }

            if llm_response.parsed_outputs:
                await self._async_log_to_cloud(
                    version_details["uuid"],
                    inputs,
                    llm_response.api_response,
                    llm_response.parsed_outputs,
                    metadata,
                )
            else:
                await self._async_log_to_cloud(
                    version_details["uuid"],
                    inputs,
                    llm_response.api_response,
                    {},
                    metadata,
                )
            return llm_response

        return async_wrapper

    def _prepare_call_args(
        self,
        prompts: List[Dict[str, str]],
        version_detail: Dict[str, Any],
        inputs: Dict[str, Any],
        kwargs,
    ):
        stringified_inputs = {key: str(value) for key, value in inputs.items()}
        messages = [
            {
                "content": prompt["content"].format(**stringified_inputs),
                "role": prompt["role"],
            }
            for prompt in prompts
        ]
        call_args = {
            "messages": messages,
            "model": version_detail["model"] if version_detail else None,
            "parsing_type": version_detail["parsing_type"] if version_detail else None,
            "output_keys": version_detail["output_keys"] if version_detail else None,
        }
        if call_args["parsing_type"] is None:
            del call_args["parsing_type"]
            del call_args["output_keys"]

        if "function_list" in kwargs:
            call_args["functions"] = kwargs["function_list"]
        return call_args

    async def _async_log_to_cloud(
        self,
        version_uuid: str,
        inputs: Optional[Dict] = None,
        api_response: Optional[ModelResponse] = None,
        parsed_outputs: Optional[Dict] = None,
        metadata: Optional[Dict] = None,
    ):
        # Perform the logging asynchronously
        if api_response:
            api_response_dict = api_response.to_dict_recursive()
            api_response_dict.update(
                {
                    "response_ms": api_response["response_ms"]
                    if "response_ms" in api_response
                    else api_response.response_ms
                }
            )
        else:
            api_response_dict = None
        res = await AsyncAPIClient.execute(
            method="POST",
            path="/log_deployment_run",
            params={
                "version_uuid": version_uuid,
            },
            json={
                "inputs": inputs,
                "api_response": api_response_dict,
                "parsed_outputs": parsed_outputs,
                "metadata": metadata,
            },
            use_cli_key=False,
        )
        if res.status_code != 200:
            print(f"[red]Failed to log to cloud: {res.json()}[/red]")
        return res

    def _sync_log_to_cloud(
        self,
        version_uuid: str,
        inputs: Optional[Dict] = None,
        api_response: Optional[ModelResponse] = None,
        parsed_outputs: Optional[Dict] = None,
        metadata: Optional[Dict] = None,
    ):
        coro = self._async_log_to_cloud(
            version_uuid,
            inputs,
            api_response,
            parsed_outputs,
            metadata,
        )
        # This is the synchronous wrapper that will wait for the async function to complete.
        try:
            # Try to get a running event loop
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # If there is no running loop, create a new one and set it as the event loop
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            result = loop.run_until_complete(coro)
            # loop.close()  # Make sure to close the loop after use
            return result

        if loop.is_running():
            # nest_asyncio.apply already done
            return loop.run_until_complete(coro)
        else:
            return loop.run_until_complete(coro)

    def run(
        self, inputs: Dict[str, Any] = {}, function_list: Optional[List[Any]] = None
    ) -> LLMResponse:
        kwargs = {"function_list": function_list} if function_list else {}
        return self._wrap_method(super().run)(inputs, **kwargs)

    def arun(
        self, inputs: Dict[str, Any] = {}, function_list: Optional[List[Any]] = None
    ) -> LLMResponse:
        kwargs = {"function_list": function_list} if function_list else {}
        return self._wrap_async_method(super().arun)(inputs, **kwargs)

    def stream(
        self, inputs: Dict[str, Any] = {}, function_list: Optional[List[Any]] = None
    ) -> Generator[LLMStreamResponse, None, None]:
        kwargs = {"function_list": function_list} if function_list else {}
        return self._wrap_gen(super().stream)(inputs, **kwargs)

    def astream(
        self,
        inputs: Optional[Dict[str, Any]] = {},
        function_list: Optional[List[Any]] = None,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        kwargs = {"function_list": function_list} if function_list else {}
        return self._wrap_async_gen(super().astream)(inputs, **kwargs)

    def run_and_parse(
        self, inputs: Dict[str, Any] = {}, function_list: Optional[List[Any]] = None
    ) -> LLMResponse:
        kwargs = {"function_list": function_list} if function_list else {}
        return self._wrap_method(super().run_and_parse)(inputs, **kwargs)

    def arun_and_parse(
        self, inputs: Dict[str, Any] = {}, function_list: Optional[List[Any]] = None
    ) -> LLMResponse:
        kwargs = {"function_list": function_list} if function_list else {}
        return self._wrap_async_method(super().arun_and_parse)(inputs, **kwargs)

    def stream_and_parse(
        self, inputs: Dict[str, Any] = {}, function_list: Optional[List[Any]] = None
    ) -> Generator[LLMStreamResponse, None, None]:
        kwargs = {"function_list": function_list} if function_list else {}
        return self._wrap_gen(super().stream_and_parse)(inputs, **kwargs)

    def astream_and_parse(
        self, inputs: Dict[str, Any] = {}, function_list: Optional[List[Any]] = None
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        kwargs = {"function_list": function_list} if function_list else {}
        return self._wrap_async_gen(super().astream_and_parse)(inputs, **kwargs)

    # def run_and_parse_function_call(
    #     self,
    #     inputs: Dict[str, Any] = {},
    #     function_list: List[Callable[..., Any]] = [],
    # ) -> Generator[str, None, None]:
    #     return self._wrap_method(super().run_and_parse_function_call)(
    #         inputs, function_list
    #     )

    # def arun_and_parse_function_call(
    #     self,
    #     inputs: Dict[str, Any] = {},
    #     function_list: List[Callable[..., Any]] = [],
    # ) -> Generator[str, None, None]:
    #     return self._wrap_async_method(super().arun_and_parse_function_call)(
    #         inputs, function_list
    #     )

    # def astream_and_parse_function_call(
    #     self,
    #     inputs: Dict[str, Any] = {},
    #     function_list: List[Callable[..., Any]] = [],
    # ) -> AsyncGenerator[str, None]:
    #     return self._wrap_async_gen(super().astream_and_parse_function_call)(
    #         inputs, function_list
    #     )



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

import asyncio
from typing import Optional
from uuid import uuid4
from dataclasses import dataclass
from websockets.exceptions import ConnectionClosedOK
from promptmodel.websocket.websocket_client import DevWebsocketClient
from promptmodel.utils.enums import LocalTask, ParsingType
from promptmodel.database.models import PromptModelVersion
from promptmodel.utils.types import FunctionSchema


@dataclass
class PromptModelInterface:
    name: str
    default_model: str = "gpt-3.5-turbo"


def get_current_weather(location: str, unit: Optional[str] = "celsius"):
    return "13 degrees celsius"


get_current_weather_desc = {
    "name": "get_current_weather",
    "description": "Get the current weather in a given location",
    "parameters": {
        "type": "object",
        "properties": {
            "location": {
                "type": "string",
                "description": "The city and state, e.g. San Francisco, CA",
            },
            "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
        },
        "required": ["location"],
    },
}


@pytest.mark.asyncio
async def test_run_model_function_call(
    mocker, websocket_client: DevWebsocketClient, mock_websocket: AsyncMock
):
    websocket_client._devapp.prompt_models = [PromptModelInterface("test_module")]
    websocket_client._devapp.samples = {
        "sample_1": {"user_message": "What is the weather like in Boston?"}
    }
    websocket_client._devapp.functions = {
        "get_current_weather": {
            "schema": FunctionSchema(**get_current_weather_desc),
            "function": get_current_weather,
        }
    }
    prompt_model_uuid = uuid4()
    param = {
        "prompt_model_uuid": prompt_model_uuid,
        "uuid": uuid4(),
        "status": "broken",
        "from_uuid": None,
        "model": "gpt-3.5-turbo",
        "parsing_type": None,
        "output_keys": None,
        "functions": [],
    }
    mock_version = PromptModelVersion(**param)
    mocker.patch(
        "promptmodel.websocket.websocket_client.get_sample_input",
        new_callable=MagicMock,
        return_value={
            "name": "sample_1",
            "contents": {"user_message": "What is the weather like in Boston?"},
        },
    )
    mocker.patch(
        "promptmodel.websocket.websocket_client.get_prompt_model_uuid",
        new_callable=MagicMock,
        return_value={"uuid": prompt_model_uuid},
    )
    mocker.patch(
        "promptmodel.websocket.websocket_client.create_prompt_model_version",
        new_callable=MagicMock,
        return_value=mock_version,
    )
    mocker.patch(
        "promptmodel.websocket.websocket_client.create_prompt", new_callable=MagicMock
    )
    mocker.patch(
        "promptmodel.websocket.websocket_client.update_prompt_model_version",
        new_callable=MagicMock,
    )
    create_run_log_mock = mocker.patch(
        "promptmodel.websocket.websocket_client.create_run_log", new_callable=MagicMock
    )

    # success case
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_PROMPT_MODEL,
            "prompt_model_name": "test_module",
            "sample_name": "sample_1",
            "prompts": [
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                    "step": 1,
                },
                {"role": "user", "content": "{user_message}", "step": 2},
            ],
            "model": "gpt-3.5-turbo",
            "uuid": None,
            "from_uuid": None,
            "parsing_type": None,
            "output_keys": None,
            "functions": ["get_current_weather"],
        },
        ws=mock_websocket,
    )
    create_run_log_mock.assert_called_once()
    _, kwargs = create_run_log_mock.call_args
    assert kwargs["function_call"] is not None, "function_call is None"
    create_run_log_mock.reset_mock()
    print(
        "======================================================================================="
    )

    # success case with no function call
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_PROMPT_MODEL,
            "prompt_model_name": "test_module",
            "sample_name": "sample_1",
            "prompts": [
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                    "step": 1,
                },
                {"role": "user", "content": "{user_message}", "step": 2},
            ],
            "model": "gpt-3.5-turbo",
            "uuid": None,
            "from_uuid": None,
            "parsing_type": None,
            "output_keys": None,
            "functions": [],
        },
        ws=mock_websocket,
    )
    create_run_log_mock.assert_called_once()
    _, kwargs = create_run_log_mock.call_args
    assert kwargs["function_call"] is None, "function_call is not None"
    create_run_log_mock.reset_mock()
    print(
        "======================================================================================="
    )

    # success case with parsing
    system_prompt_with_format = """
You are a helpful assistant.

This is your output format. Keep the string between < type=< >>, </ > as it is.
<temperature type=<float>>
(value here)
</temperature>
    """
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_PROMPT_MODEL,
            "prompt_model_name": "test_module",
            "sample_name": "sample_1",
            "prompts": [
                {"role": "system", "content": system_prompt_with_format, "step": 1},
                {"role": "user", "content": "{user_message}", "step": 2},
            ],
            "model": "gpt-3.5-turbo",
            "uuid": None,
            "from_uuid": None,
            "parsing_type": ParsingType.HTML.value,
            "output_keys": ["temperature"],
            "functions": ["get_current_weather"],
        },
        ws=mock_websocket,
    )
    create_run_log_mock.assert_called_once()
    _, kwargs = create_run_log_mock.call_args
    assert kwargs["function_call"] is not None, "function_call is None"
    create_run_log_mock.reset_mock()
    print(
        "======================================================================================="
    )

    # early failed case due to function not existing
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_PROMPT_MODEL,
            "prompt_model_name": "test_module",
            "sample_name": "sample_1",
            "prompts": [
                {
                    "role": "system",
                    "content": "You are a helpful assistant.",
                    "step": 1,
                },
                {"role": "user", "content": "{user_message}", "step": 2},
            ],
            "model": "gpt-3.5-turbo",
            "uuid": None,
            "from_uuid": None,
            "parsing_type": None,
            "output_keys": None,
            "functions": ["get_weather"],
        },
        ws=mock_websocket,
    )
    create_run_log_mock.assert_not_called()



import os
import time
import asyncio
import typer
import json
import sys
import importlib
import signal
from importlib import resources
from requests import request
from typing import Callable, Dict, Any, List

import webbrowser
import inspect

from rich import print
from InquirerPy import inquirer
from watchdog.observers import Observer

from promptmodel import DevApp
from promptmodel.apis.base import APIClient
from promptmodel.constants import ENDPOINT_URL, WEB_CLIENT_URL
from promptmodel.cli.commands.init import init as promptmodel_init
from promptmodel.cli.utils import get_org, get_project
from promptmodel.cli.signal_handler import dev_terminate_signal_handler
import promptmodel.utils.logger as logger
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.utils.crypto import generate_api_key, encrypt_message
from promptmodel.utils.enums import PromptModelVersionStatus, ChangeLogAction
from promptmodel.websocket import DevWebsocketClient, CodeReloadHandler
from promptmodel.database.orm import initialize_db
from promptmodel.database.models import PromptModelVersion
from promptmodel.database.crud import (
    create_prompt_models,
    create_prompt_model_versions,
    create_prompts,
    create_run_logs,
    list_prompt_models,
    hide_prompt_model_not_in_code,
    update_samples,
    update_prompt_model_uuid,
    rename_prompt_model,
)


def dev():
    """Creates a new prompt development environment, and opens up DevApp in the browser."""
    upsert_config({"initializing": True}, "dev_branch")
    signal.signal(signal.SIGINT, dev_terminate_signal_handler)
    promptmodel_init(from_cli=False)

    _devapp_filename, devapp_instance_name = "promptmodel_dev:app".split(":")

    # Init local database & open
    initialize_db()

    config = read_config()

    devapp_module = importlib.import_module(_devapp_filename)
    devapp_instance: DevApp = getattr(devapp_module, devapp_instance_name)

    if "name" not in config["dev_branch"]:
        org = get_org(config)
        project = get_project(config=config, org=org)
        validate_branch_name = lambda name: APIClient.execute(
            method="GET", path="/check_dev_branch_name", params={"name": name}
        ).json()
        branch_name = inquirer.text(
            message="Enter a development branch name:",
            validate=lambda x: " " not in x and validate_branch_name(x),
            invalid_message="Branch name already exists or contains spaces.",
        ).execute()

        print("\nCreating local development branch...")
        created_dev_branch_row = APIClient.execute(
            method="POST",
            path="/create_dev_branch",
            params={"name": branch_name, "project_uuid": project["uuid"]},
        )
        upsert_config(
            {
                "name": branch_name,
                "project": project,
                "org": org,
                "uuid": created_dev_branch_row.json()["uuid"],
                "project_version": "0.0.0",
            },
            section="dev_branch",
        )

        # connect
        res = APIClient.execute(
            method="POST",
            path="/connect_cli_dev",
            params={"project_uuid": project["uuid"], "branch_name": branch_name},
        )
        if res.status_code != 200:
            print(f"Error: {res.json()['detail']}")
            return

        # fetch current project status
        project_status = APIClient.execute(
            method="GET",
            path="/pull_project",
            params={"project_uuid": project["uuid"]},
        ).json()
        # save project version
        upsert_config(
            {"project_version": project_status["project_version"]}, section="dev_branch"
        )

        local_prompt_models = devapp_instance.prompt_models
        local_prompt_model_names = [x.name for x in local_prompt_models]

        # save prompt_models
        for prompt_model in project_status["prompt_models"]:
            prompt_model["is_deployed"] = True
            if prompt_model["name"] in local_prompt_model_names:
                prompt_model["used_in_code"] = True
            else:
                prompt_model["used_in_code"] = False

        create_prompt_models(project_status["prompt_models"])

        # save prompt_model_versions
        for version in project_status["prompt_model_versions"]:
            version["status"] = PromptModelVersionStatus.CANDIDATE.value
        create_prompt_model_versions(project_status["prompt_model_versions"])
        # save prompts
        create_prompts(project_status["prompts"])
        # save run_logs
        create_run_logs(project_status["run_logs"])

        # create prompt_models from code in local DB
        project_prompt_model_names = [
            x["name"] for x in project_status["prompt_models"]
        ]
        only_in_local = list(
            set(local_prompt_model_names) - set(project_prompt_model_names)
        )
        only_in_local_prompt_models = [
            {"name": x, "project_uuid": project["uuid"]} for x in only_in_local
        ]
        create_prompt_models(only_in_local_prompt_models)

    else:
        org = config["dev_branch"]["org"]
        project = config["dev_branch"]["project"]
        branch_name = config["dev_branch"]["name"]

        res = APIClient.execute(
            method="POST",
            path="/connect_cli_dev",
            params={"project_uuid": project["uuid"], "branch_name": branch_name},
        )
        if res.status_code != 200:
            print(f"Error: {res.json()['detail']}")
            return

        # Get changelog & update DB
        # fetch current project status
        project_status = APIClient.execute(
            method="GET",
            path="/pull_project",
            params={"project_uuid": project["uuid"]},
        ).json()

        changelogs = APIClient.execute(
            method="GET",
            path="/get_changelog",
            params={
                "project_uuid": project["uuid"],
                "local_project_version": config["dev_branch"]["project_version"],
                "levels": [1, 2],
            },
        ).json()
        local_code_prompt_model_name_list = [
            x.name for x in devapp_instance.prompt_models
        ]
        res = update_by_changelog(
            changelogs, project_status, local_code_prompt_model_name_list
        )
        if res is False:
            print("Update Dev Stopped.")
            upsert_config({"online": False}, section="dev_branch")
            return

        # local_code_prompt_model_name_list 에 없는 prompt_model 의 used_in_code=False 설정
        hide_prompt_model_not_in_code(local_code_prompt_model_name_list)

        # 새로 생긴 prompt_model DB에 생성
        local_db_prompt_model_names = [x["name"] for x in list_prompt_models()]
        only_in_local = list(
            set(local_code_prompt_model_name_list) - set(local_db_prompt_model_names)
        )
        only_in_local_prompt_models = [
            {"name": x, "project_uuid": project["uuid"]} for x in only_in_local
        ]
        create_prompt_models(only_in_local_prompt_models)

    dev_url = f"{WEB_CLIENT_URL}/org/{org['slug']}/projects/{project['uuid']}/dev/{branch_name}"

    # Open websocket connection to backend server
    dev_websocket_client = DevWebsocketClient(_devapp=devapp_instance)

    import threading

    reloader_thread = threading.Thread(
        target=start_code_reloader,
        args=(_devapp_filename, devapp_instance_name, dev_websocket_client),
    )
    reloader_thread.daemon = True  # Set the thread as a daemon
    reloader_thread.start()

    print(
        f"\nOpening [violet]PromptModel[/violet] prompt engineering environment with the following configuration:\n"
    )
    print(f"📌 Organization: [blue]{org['name']}[/blue]")
    print(f"📌 Project: [blue]{project['name']}[/blue]")
    print(f"📌 Local development branch: [green][bold]{branch_name}[/bold][/green]")
    print(
        f"\nIf browser doesn't open automatically, please visit [link={dev_url}]{dev_url}[/link]"
    )
    webbrowser.open(dev_url)

    upsert_config({"online": True, "initializing": False}, section="dev_branch")
    # save samples to local DB
    update_samples(devapp_instance.samples)

    # Open Websocket
    asyncio.run(
        dev_websocket_client.connect_to_gateway(
            project_uuid=project["uuid"],
            dev_branch_name=branch_name,
            cli_access_header=APIClient._get_headers(),
        )
    )


app = typer.Typer(invoke_without_command=True, callback=dev)


def start_code_reloader(_devapp_filename, devapp_instance_name, dev_websocket_client):
    event_handler = CodeReloadHandler(
        _devapp_filename, devapp_instance_name, dev_websocket_client
    )
    observer = Observer()
    observer.schedule(event_handler, path=".", recursive=True)
    observer.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def update_by_changelog(
    changelogs: List[Dict],
    project_status: dict,
    local_code_prompt_model_name_list: List[str],
):
    """Update Local DB by changelog"""
    local_db_prompt_model_list: list = list_prompt_models()  # {"name", "uuid"}

    for changelog in changelogs:
        level: int = changelog["level"]
        logs = changelog["logs"]
        if level == 1:
            for log in logs:
                subject = log["subject"]
                action: str = log["action"]
                if subject == "prompt_model":
                    local_db_prompt_model_list = update_prompt_model_changelog(
                        action=action,
                        project_status=project_status,
                        uuid_list=log["identifiers"],
                        local_db_prompt_model_list=local_db_prompt_model_list,
                        local_code_prompt_model_name_list=local_code_prompt_model_name_list,
                    )
                elif subject == "prompt_model_version":
                    (
                        is_success,
                        local_db_prompt_model_list,
                    ) = update_prompt_model_version_changelog(
                        action=action,
                        project_status=project_status,
                        uuid_list=log["identifiers"],
                        local_db_prompt_model_list=local_db_prompt_model_list,
                        local_code_prompt_model_name_list=local_code_prompt_model_name_list,
                    )
                    if not is_success:
                        return False
                else:
                    pass
            previous_version_levels = changelog["previous_version"].split(".")
            current_version_levels = [
                str(int(previous_version_levels[0]) + 1),
                "0",
                "0",
            ]
            current_version = ".".join(current_version_levels)
        elif level == 2:
            for log in logs:
                subject = log["subject"]
                action: str = log["action"]
                uuid_list: list = log["identifiers"]
                if subject == "prompt_model_version":
                    local_db_prompt_model_list = update_prompt_model_version_changelog(
                        action=action,
                        project_status=project_status,
                        uuid_list=log["identifiers"],
                        local_db_prompt_model_list=local_db_prompt_model_list,
                        local_code_prompt_model_name_list=local_code_prompt_model_name_list,
                    )
                else:
                    pass
            previous_version_levels = changelog["previous_version"].split(".")
            current_version_levels = [
                previous_version_levels[0],
                str(int(previous_version_levels[1]) + 1),
                "0",
            ]
            current_version = ".".join(current_version_levels)
        else:
            previous_version_levels = changelog["previous_version"].split(".")
            current_version_levels = [
                previous_version_levels[0],
                previous_version_levels[1],
                str(int(previous_version_levels[2]) + 1),
            ]
            current_version = ".".join(current_version_levels)

        upsert_config({"project_version": current_version}, section="dev_branch")
    return True


def update_prompt_model_changelog(
    action: ChangeLogAction,
    project_status: dict,
    uuid_list: List[str],
    local_db_prompt_model_list: List[Dict],
    local_code_prompt_model_name_list: List[str],
):
    if action == ChangeLogAction.ADD.value:
        prompt_model_list = [
            x for x in project_status["prompt_models"] if x["uuid"] in uuid_list
        ]
        for prompt_model in prompt_model_list:
            local_db_prompt_model_name_list = [
                x["name"] for x in local_db_prompt_model_list
            ]

            if prompt_model["name"] not in local_db_prompt_model_name_list:
                # IF prompt_model not in Local DB
                if prompt_model["name"] in local_code_prompt_model_name_list:
                    # IF prompt_model in Local Code
                    prompt_model["used_in_code"] = True
                    prompt_model["is_deployed"] = True
                    create_prompt_models([prompt_model])
                else:
                    prompt_model["used_in_code"] = False
                    prompt_model["is_deployed"] = True
                    create_prompt_models([prompt_model])
            else:
                local_db_prompt_model = [
                    x
                    for x in local_db_prompt_model_list
                    if x["name"] == prompt_model["name"]
                ][0]
                if local_db_prompt_model["is_deployed"] is False:
                    print(
                        "Creation of promptmodel with identical name was detected in local & deployment."
                    )
                    check_same = inquirer.confirm(
                        message="Are they same promptmodel? [y/n]", default=False
                    ).execute()
                    # check_same = input()
                    # if check_same == "y":
                    #     check_same = True
                    # else:
                    #     check_same = False

                    if not check_same:
                        # rename & change name in Local DB
                        print(
                            f"Please rename local promptmodel {prompt_model['name']} to continue"
                        )
                        validate_new_promptmodel_name = (
                            lambda name: name not in local_code_prompt_model_name_list
                            and name not in local_db_prompt_model_name_list
                        )
                        new_model_name = inquirer.text(
                            message="Enter the new promptmodel name:",
                            validate=lambda x: validate_new_promptmodel_name(x),
                            invalid_message="promptmodel name already exists.",
                        ).execute()
                        # new_model_name = input()
                        rename_prompt_model(
                            local_db_prompt_model["uuid"], new_model_name
                        )

                        print("We changed the name of promptmodel in local DB.")
                        print(
                            f"Please change the name of promptmodel '{prompt_model['name']}' in your project code and restart."
                        )
                        # dev 꺼버리고, 수정하고 다시 키라고 명령.
                        return False, local_db_prompt_model_list
                    update_prompt_model_uuid(
                        local_db_prompt_model["uuid"], prompt_model["uuid"]
                    )
                    local_db_prompt_model_list: list = list_prompt_models()
    else:
        # TODO: add code DELETE, CHANGE, FIX later
        pass
    return True, local_db_prompt_model_list


def update_prompt_model_version_changelog(
    action: ChangeLogAction,
    project_status: dict,
    uuid_list: List[str],
    local_db_prompt_model_list: List[Dict],
    local_code_prompt_model_name_list: List[str],
) -> List[Dict[str, Any]]:
    if action == ChangeLogAction.ADD.value:
        # find prompt_model_version in project_status['prompt_model_versions'] where uuid in uuid_list
        prompt_model_version_list_in_changelog = [
            x for x in project_status["prompt_model_versions"] if x["uuid"] in uuid_list
        ]

        # check if prompt_model_version['uuid'] is in local_db_prompt_model_list
        local_db_version_uuid_list = [
            str(x.uuid) for x in list(PromptModelVersion.select())
        ]
        version_list_to_update = [
            x
            for x in prompt_model_version_list_in_changelog
            if x["uuid"] not in local_db_version_uuid_list
        ]
        version_uuid_list_to_update = [x["uuid"] for x in version_list_to_update]

        # find prompts and run_logs to update
        prompts_to_update = [
            x
            for x in project_status["prompts"]
            if x["version_uuid"] in version_uuid_list_to_update
        ]
        run_logs_to_update = [
            x
            for x in project_status["run_logs"]
            if x["version_uuid"] in version_uuid_list_to_update
        ]

        for prompt_model_version in version_list_to_update:
            prompt_model_version["status"] = PromptModelVersionStatus.CANDIDATE.value

        create_prompt_model_versions(version_list_to_update)
        create_prompts(prompts_to_update)
        create_run_logs(run_logs_to_update)

        return local_db_prompt_model_list
    else:
        pass



from enum import Enum
import datetime
import json

from uuid import uuid4
from peewee import (
    CharField,
    DateTimeField,
    IntegerField,
    ForeignKeyField,
    BooleanField,
    UUIDField,
    TextField,
    AutoField,
    FloatField,
    Check,
)

from promptmodel.database.config import BaseModel
from promptmodel.utils.enums import PromptModelVersionStatus, ParsingType


class JSONField(TextField):
    def db_value(self, value):
        return json.dumps(value)

    def python_value(self, value):
        return json.loads(value)


class PromptModel(BaseModel):
    uuid = UUIDField(unique=True, default=uuid4)
    created_at = DateTimeField(default=datetime.datetime.now)
    project_uuid = UUIDField()
    name = CharField()
    used_in_code = BooleanField(default=True)
    is_deployed = BooleanField(
        default=False
    )  # mark if the module is pushed to the cloud


class PromptModelVersion(BaseModel):
    uuid = UUIDField(unique=True, default=uuid4)
    created_at = DateTimeField(default=datetime.datetime.now)
    from_uuid = UUIDField(null=True)
    prompt_model_uuid = ForeignKeyField(
        PromptModel,
        field=PromptModel.uuid,
        backref="versions",
        on_delete="CASCADE",
    )
    status = CharField(
        constraints=[
            Check(
                f"status IN ('{PromptModelVersionStatus.BROKEN.value}', '{PromptModelVersionStatus.WORKING.value}', '{PromptModelVersionStatus.CANDIDATE.value}')"
            )
        ]
    )
    model = CharField()
    version = IntegerField(null=True)
    is_published = BooleanField(default=False)
    is_deployed = BooleanField(default=False)
    parsing_type = CharField(
        null=True,
        default=None,
        constraints=[
            Check(
                f"parsing_type IN ('{ParsingType.COLON.value}', '{ParsingType.SQUARE_BRACKET.value}', '{ParsingType.DOUBLE_SQUARE_BRACKET.value}')"
            )
        ],
    )
    output_keys = JSONField(null=True, default=None)
    functions = JSONField(default=[])


class Prompt(BaseModel):
    id = AutoField()
    created_at = DateTimeField(default=datetime.datetime.now)
    version_uuid = ForeignKeyField(
        PromptModelVersion,
        field=PromptModelVersion.uuid,
        backref="prompts",
        on_delete="CASCADE",
    )
    role = CharField()
    step = IntegerField()
    content = TextField()


class RunLog(BaseModel):
    id = AutoField()
    created_at = DateTimeField(default=datetime.datetime.now)
    version_uuid = ForeignKeyField(
        PromptModelVersion,
        field=PromptModelVersion.uuid,
        backref="run_logs",
        on_delete="CASCADE",
    )
    inputs = JSONField(null=True, default={})
    raw_output = TextField()
    parsed_outputs = JSONField(null=True, default={})
    run_from_deployment = BooleanField(default=False)
    function_call = JSONField(null=True, default={})


class SampleInputs(BaseModel):
    id = AutoField()
    created_at = DateTimeField(default=datetime.datetime.now)
    name = TextField(unique=True)
    contents = JSONField()


class DeployedPromptModel(BaseModel):
    uuid = UUIDField(unique=True, default=uuid4)
    name = CharField()


class DeployedPromptModelVersion(BaseModel):
    uuid = UUIDField(unique=True, default=uuid4)
    from_uuid = UUIDField(null=True)
    prompt_model_uuid = ForeignKeyField(
        DeployedPromptModel,
        field=DeployedPromptModel.uuid,
        backref="versions",
        on_delete="CASCADE",
    )
    model = CharField()
    is_published = BooleanField(default=False)
    is_ab_test = BooleanField(default=False)
    ratio = FloatField(null=True)
    parsing_type = CharField(
        null=True,
        default=None,
        constraints=[
            Check(
                f"parsing_type IN ('{ParsingType.COLON.value}', '{ParsingType.SQUARE_BRACKET.value}', '{ParsingType.DOUBLE_SQUARE_BRACKET.value}')"
            )
        ],
    )
    output_keys = JSONField(null=True, default=None)
    functions = JSONField(default=[])


class DeployedPrompt(BaseModel):
    id = AutoField()
    version_uuid = ForeignKeyField(
        DeployedPromptModelVersion,
        field=DeployedPromptModelVersion.uuid,
        backref="prompts",
        on_delete="CASCADE",
    )
    role = CharField()
    step = IntegerField()
    content = TextField()



import os

from .models import (
    PromptModel,
    PromptModelVersion,
    Prompt,
    RunLog,
    SampleInputs,
    DeployedPromptModel,
    DeployedPromptModelVersion,
    DeployedPrompt,
)
from .config import db


def initialize_db():
    if not os.path.exists("./.promptmodel"):
        os.mkdir("./.promptmodel")
    # Check if db connection exists
    if db.is_closed():
        db.connect()
        with db.atomic():
            if not PromptModel.table_exists():
                db.create_tables(
                    [
                        PromptModel,
                        PromptModelVersion,
                        Prompt,
                        RunLog,
                        SampleInputs,
                        DeployedPromptModel,
                        DeployedPromptModelVersion,
                        DeployedPrompt,
                    ]
                )
        db.close()



import os
import sys
import yaml
import asyncio
from datetime import datetime
from threading import Thread
from typing import Any, Dict, Tuple, List, Union, Optional, Coroutine
from litellm import token_counter

from promptmodel.apis.base import AsyncAPIClient
from promptmodel.database.crud import (
    get_latest_version_prompts,
    get_deployed_prompts,
)
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.utils import logger
from promptmodel.utils.random_utils import select_version
from promptmodel.promptmodel_init import CacheManager, update_deployed_db


async def fetch_prompts(name) -> Tuple[List[Dict[str, str]], Dict[str, Any]]:
    """fetch prompts.

    Args:
        name (str): name of promtpmodel

    Returns:
        Tuple[List[Dict[str, str]], Optional[Dict[str, Any]]]: (prompts, version_detail)
    """
    # Check dev_branch activate
    config = read_config()
    if "dev_branch" in config and config["dev_branch"]["initializing"] == True:
        return [], {}
    elif "dev_branch" in config and config["dev_branch"]["online"] == True:
        # get prompt from local DB
        prompt_rows, version_detail = get_latest_version_prompts(name)
        if prompt_rows is None:
            return [], {}
        return [
            {"role": prompt.role, "content": prompt.content} for prompt in prompt_rows
        ], version_detail
    else:
        if (
            "project" in config
            and "use_cache" in config["project"]
            and config["project"]["use_cache"] == True
        ):
            cache_manager = CacheManager()
            # call update_local API in background task
            cache_update_thread = Thread(
                target=cache_manager.cache_update_background_task, args=(config,)
            )
            cache_update_thread.daemon = True
            cache_update_thread.start()

            # get prompt from local DB by ratio
            prompt_rows, version_detail = get_deployed_prompts(name)
            if prompt_rows is None:
                return [], {}

            return [
                {"role": prompt.role, "content": prompt.content}
                for prompt in prompt_rows
            ], version_detail

        else:
            try:
                prompts_data = await AsyncAPIClient.execute(
                    method="GET",
                    path="/fetch_published_prompt_model_version",
                    params={"prompt_model_name": name},
                    use_cli_key=False,
                )
                prompts_data = prompts_data.json()
            except Exception as e:
                raise e
            prompt_model_versions = prompts_data["prompt_model_versions"]
            prompts = prompts_data["prompts"]
            for version in prompt_model_versions:
                if version["is_published"] is True:
                    version["ratio"] = 1.0
            selected_version = select_version(prompt_model_versions)

            prompt_rows = list(
                filter(
                    lambda prompt: str(prompt["version_uuid"])
                    == str(selected_version["uuid"]),
                    prompts,
                )
            )
            # sort prompt_rows by step
            prompt_rows = sorted(prompt_rows, key=lambda prompt: prompt["step"])

            version_detail = {
                "model": selected_version["model"],
                "uuid": selected_version["uuid"],
                "parsing_type": selected_version["parsing_type"],
                "output_keys": selected_version["output_keys"],
            }

            if prompt_rows is None:
                return [], {}

            return [
                {"role": prompt["role"], "content": prompt["content"]}
                for prompt in prompt_rows
            ], version_detail


def set_inputs_to_prompts(inputs: Dict[str, Any], prompts: List[Dict[str, str]]):
    messages = [
        {"content": prompt["content"].format(**inputs), "role": prompt["role"]}
        for prompt in prompts
    ]
    return messages


def num_tokens_for_messages(
    messages: List[Dict[str, str]], model: str = "gpt-3.5-turbo-0613"
) -> int:
    tokens_per_message = 0
    tokens_per_name = 0
    if model.startswith("gpt-3.5-turbo"):
        tokens_per_message = 4
        tokens_per_name = -1

    if model.startswith("gpt-4"):
        tokens_per_message = 3
        tokens_per_name = 1

    if model.endswith("-0613") or model == "gpt-3.5-turbo-16k":
        tokens_per_message = 3
        tokens_per_name = 1
    sum = 0
    sum = token_counter(model=model, messages=messages)
    for message in messages:
        sum += tokens_per_message
        if "name" in message:
            sum += tokens_per_name
    return sum


def num_tokens_from_functions_input(
    functions: List[Any], model="gpt-3.5-turbo-0613"
) -> int:
    """Return the number of tokens used by a list of functions."""

    num_tokens = 0
    for function in functions:
        function_tokens = token_counter(model=model, text=function["name"])
        function_tokens += token_counter(model=model, text=function["description"])

        if "parameters" in function:
            parameters = function["parameters"]
            if "properties" in parameters:
                for properties_key in parameters["properties"]:
                    function_tokens += token_counter(model=model, text=properties_key)
                    v = parameters["properties"][properties_key]
                    for field in v:
                        if field == "type":
                            function_tokens += 2
                            function_tokens += token_counter(
                                model=model, text=v["type"]
                            )
                        elif field == "description":
                            function_tokens += 2
                            function_tokens += token_counter(
                                model=model, text=v["description"]
                            )
                        elif field == "enum":
                            function_tokens -= 3
                            for o in v["enum"]:
                                function_tokens += 3
                                function_tokens += token_counter(model=model, text=o)
                        else:
                            print(f"Warning: not supported field {field}")
                function_tokens += 11

        num_tokens += function_tokens

    num_tokens += 12
    return num_tokens


def num_tokens_from_function_call_output(
    function_call_output: Dict[str, str] = {}, model="gpt-3.5-turbo-0613"
) -> int:
    num_tokens = 1
    num_tokens += token_counter(model=model, text=function_call_output["name"])
    if "arguments" in function_call_output:
        num_tokens += token_counter(model=model, text=function_call_output["arguments"])
    return num_tokens


import asyncio


def run_async_in_sync(coro: Coroutine):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:  # No running loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(coro)
        # loop.close()
        return result

    if loop.is_running():
        # nest_asyncio.apply already done
        return loop.run_until_complete(coro)
    else:
        return loop.run_until_complete(coro)



from typing import (
    List,
    Dict,
    Tuple,
    Any,
    Generator,
    Optional,
    AsyncGenerator,
    Callable,
    Union,
)
from pydantic import BaseModel
from litellm import ModelResponse as LiteLLMModelResponse


class ModelResponse(LiteLLMModelResponse):
    def __init__(self, **params):
        super(ModelResponse, self).__init__(**params)


class LLMResponse:
    def __init__(
        self,
        api_response: Optional[ModelResponse] = None,
        raw_output: Optional[str] = None,
        parsed_outputs: Optional[Dict[str, str]] = None,
        error: Optional[bool] = None,
        error_log: Optional[str] = None,
        function_call: Optional[Dict[str, Any]] = None,
    ):
        self.api_response = api_response
        self.raw_output = raw_output
        self.parsed_outputs = parsed_outputs
        self.error = error
        self.error_log = error_log
        self.function_call = function_call


class LLMStreamResponse:
    def __init__(
        self,
        api_response: Optional[ModelResponse] = None,
        raw_output: Optional[str] = None,
        parsed_outputs: Optional[Dict[str, str]] = None,
        error: Optional[bool] = None,
        error_log: Optional[str] = None,
        function_call: Optional[Dict[str, Any]] = None,
    ):
        self.api_response = api_response
        self.raw_output = raw_output
        self.parsed_outputs = parsed_outputs
        self.error = error
        self.error_log = error_log
        self.function_call = function_call


class FunctionSchema(BaseModel):
    """
    {
            "name": str,
            "description": Optional[str],
            "parameters": {
                "type": "object",
                "properties": {
                    "argument_name": {
                        "type": str,
                        "description": Optional[str],
                        "enum": Optional[List[str]]
                    },
                },
                "required": Optional[List[str]],
            },
        }
    """

    class _Parameters(BaseModel):
        class _Properties(BaseModel):
            type: str
            description: Optional[str] = ""
            enum: Optional[List[str]] = []

        type: str = "object"
        properties: Dict[str, _Properties] = {}
        required: Optional[List[str]] = []

    name: str
    description: Optional[str] = None
    parameters: _Parameters
